"""This module will be for reproducing Tom Lampo's results on MSDs of colloidal
particles in confinement. Should ask Phillip Geissler's student about in what
cases in found normal diffusivity."""
import math
from ..mc import sphere_insertion

max_sphere_density = math.pi/(3*math.sqrt(2))
max_circle_density = math.pi/(2*math.sqrt(3))

def confinement_force(sphere1, containing_sphere):
    """Returns the force exerted on sphere1 by its proximity to the confinement
    sphere."""
    pass

def space_exclusion_force(sphere1, sphere2):
    """Returns the force exerted on each of two spheres due to their
    interactions with each other.

    After reading a lot about how people work with confinement, I've got the
    intuition that it's standard to arbitrarily choose among the following
    depending on what makes your particular problem more tractable:

    1) full lennard-jones
    2) lennard-jones, but only the (a/r)^12 bit without the attractive bit
    3) something "slightly harder" than gaussian, in order to increase time
    step, e.g.
    \frac{A}{3 b \delta^2}(h-\delta)^3, where \delta = b*(L/N)^{1/2}/2, A
    arbitrarily ~25k_BT"""
    pass

def bd_step(dt, spheres, containing_sphere):
    """Moves the spheres forward a time step.

    TODO: Check if returning the list of new spheres is faster."""
    pass

def bin_from_scratch():
    #TODO use histogramdd
    pass

def bd_sim(nstep, dt, target_density, sphere_radii, confinement_radius,
           bins_per_side, num_init_steps=10000):
    """Run a simulation of the spheres of sizes requested."""
    # the bins will tile the bounding box of the confining sphere, so the overall
    # bounding box has
    side_length = 2*confinement_radius
    # and each little bin has a side_length given by
    bin_length = side_length/bins_per_side
    bin_volume = bin_length**3
    # max_spheres_per_bin*volume_of_sphere = bin_volume*max_sphere_density
    sphere_volume = (4/3)*math.pi*(sphere_radii**3)
    max_spheres_in_bin = math.ceil(max_sphere_density*bin_volume/sphere_volume)
    num_bins = bins_per_side**3
    sphere_centers = np.zeros((num_bins, max_spheres_in_bin, 3))
    num_in_bin = np.zeros((num_bins,))
    # run num_init_steps MC steps to get spheres in reasonable positions
    init_spheres = mc.sphere_insertion.step_away_mc(num_init_steps, target_density, sphere_radii, confinement_radius)
    #TODO put spheres into bins
    for i in range(nstep):
        #TODO ipmlement bin-based algorithm for fast simulation




import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm, colors
from  scipy.special import sph_harm as Y
import pdb

def plot_surf(fun, title="",Rmax=None, limits=3.5):
    PHI, THETA = np.mgrid[0:2*np.pi:200j, 0:np.pi:100j] #arrays of angular variables
    R = np.abs(fun(PHI, THETA)) #Array with the absolute values of Ylm
    #Now we convert to cartesian coordinates
    # for the 3D representation
    X = R * np.sin(THETA) * np.cos(PHI)
    Y = R * np.sin(THETA) * np.sin(PHI)
    Z = R * np.cos(THETA)
    
    
    if Rmax is None:
        N = R/R.max()    # Normalize R for the plot colors to cover the entire range of colormap.
    else:
        N = R/Rmax
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d'), figsize=(12,10))
    im = ax.plot_surface(X, Y, Z, rstride=1, cstride=1, facecolors=cm.jet(N))
    ax.set_title(title, fontsize=20)
    m = cm.ScalarMappable(cmap=cm.jet)
    m.set_array(R)    # Assign the unnormalized data array to the mappable
                      #so that the scale corresponds to the values of R
    fig.colorbar(m, shrink=0.8)

    Xb = limits*np.mgrid[-1:2:2,-1:2:2,-1:2:2][0].flatten() 
    Yb = limits*np.mgrid[-1:2:2,-1:2:2,-1:2:2][1].flatten() 
    Zb = limits*np.mgrid[-1:2:2,-1:2:2,-1:2:2][2].flatten() 
    # Comment or uncomment following both lines to test the fake bounding box:
    for xb, yb, zb in zip(Xb, Yb, Zb):
       ax.plot([xb], [yb], [zb], 'w')
    plt.axis('off')
    plt.savefig(title+".png")
    plt.show()



def YR(m, ell, phi, theta):
    '''
    Real spherical harmonic function.
    '''
    if m<0:
        out = (1.0j/np.sqrt(2))*(Y(m, ell, phi, theta)
                               - (-1)**m * Y(-m, ell, phi, theta))
    if m==0:
        out = Y(m, ell, phi, theta)
    if m>0:
        out = (1.0/np.sqrt(2))*(Y(-m, ell, phi, theta) 
                             + (-1)**m * Y(m, ell, phi, theta))

    if np.isnan(np.real(out)):
        pdb.set_trace()

    return np.real(out)



fields = [[{"ell":2,"m":0,"phi":1.0}]
         , [{"ell":2,"m":0,"phi":1.0},{"ell":2,"m":0,"phi":0.4}]
         , [{"ell":2,"m":0,"phi":1.0},{"ell":2,"m":1,"phi":0.5}]
         , [{"ell":2,"m":0,"phi":1.0},{"ell":2,"m":-1,"phi":0.5}]
         , [{"ell":2,"m":0,"phi":1.0},{"ell":2,"m":2,"phi":0.9}]
         , [{"ell":2,"m":0,"phi":1.0},{"ell":2,"m":-2,"phi":0.9}] ]
titles = ["main", "m0", "m1", "m_minus_1", "m2", "m_minus_2"]
for ii in range(6):
    title=titles[ii]
    field = fields[ii]

    a=1.0
    def fun(phi, theta):
        if hasattr(phi,"__len__"):
            out = phi*0
            for  ii in range(len(phi)):
                out[ii] = fun(phi[ii], theta[ii])
            return out
        temp=0.0
        for X in field:
            temp += a*X["phi"]*YR(X["m"], X["ell"], phi, theta)
        if np.isnan(temp):
           pdb.set_trace() 
        return np.exp(temp)
    
    plot_surf(fun, Rmax=2.3, title=title)



import numpy as np
import pdb

from .stonefence import precalculate_data as calc_G_set
from .stonefence import ORDER_L
from .yyyreal import YYY
from .invlaplace import invLaplace, make_path, plot_invLaplace_path, invLaplace_path, plot_path
from .rigidrod import rodFrank
from .rigidrod import RR_datapoint

def get_first_pole(gamma, tol=0.000001):
    '''Search the real axes for polls of G[0][0,0]. Returns highest real value.

    Args:
        gamma (float): Field strength.
        tol (float): absolute tolerance of result
    '''
    nninitial=50
    if gamma<=0:
        return 0.0
    if gamma<200:
        p_initial = np.linspace(-0.01, gamma/1.99, nninitial)
    else:
        p_initial = np.linspace(-0.01, gamma, nninitial)
    G_initial = np.zeros(nninitial)
    nums = np.zeros(nninitial)
    for ii, p in enumerate(p_initial):
        G = calc_G_set(p, gamma, m_values=[0])['Gmll']
        nums[ii] = np.real(G[0][0,0])
    i_above = nninitial-1
    if nums[i_above]<0.0:
        print("gamma="+str(gamma))
        raise ValueError("Out of range! Search larger area")
    for ii in range(nninitial-1):
        if nums[i_above-1] >= 0.0:
            i_above=i_above-1
        else:
            break
    lower = p_initial[i_above-1]
    upper = p_initial[i_above]
    while True:
        if upper-lower<tol:
            return (upper+lower)*0.5
        p = (upper+lower)*0.5
        G = calc_G_set(p, gamma, m_values=[0])['Gmll']
        test = np.real(G[0][0,0])
        if test>=0:
            upper=p
        else:
            lower=p
            
            
e0 = np.zeros(ORDER_L); e0[0] = 1.0 # Unit vector
def GJG(G):
    J1 = YYY(0,2,0,0, ORDER_L=ORDER_L)
    return e0@G[0]@J1@G[0]@e0

def GJGJG(G, l1, l2, m1, m2):
    if m1 != m2:
        return 0.0
    J1 = YYY(0, l1, m1, m1, ORDER_L=ORDER_L)
    J2 = YYY(m1, l2, m2, 0, ORDER_L=ORDER_L)
    return e0@G[0]@J1@G[m1]@J2@G[0]@e0

def GJGJGJG(G, l1, l2, m1, m2, mkmk):
    (mk1, mk2) = mkmk
    J1 = YYY(0, l1, m1, m1, ORDER_L=ORDER_L)
    J4 = YYY(m2, l2, m2, 0, ORDER_L=ORDER_L)

    out=0.0+0.0j
    for M in {m1+mk1, m1-mk1, mk1-m1}:
        if M not in {m2+mk2, m2-mk2, mk2-m2}:
            continue
        J2 = YYY(m1, 1, mk1, M, ORDER_L=ORDER_L)
        J3 = YYY(M, 1, mk2, m2, ORDER_L=ORDER_L)
        out = out + e0@G[0]@J1@G[m1]@J2@G[M]@J3@G[m2]@J4@G[0]@e0
    return out

def get_a(gamma, N, gammaNmax=None):
    """
    Maier-Saupe parameter need to generate field strength gamma.

    Args:
        gamma (real): Field strength in kT's per Kuhn length of chain
        N (real): How many Kuhn lengths long the polymers are

    Note: when N=0.0, gamma is assume to be gamma*N, that is kT's per chain.
    """
    if N==0.0:
        return {'aLAf': RR_datapoint(gamma)['aLAf']}
    args={'factor':400,
          'scale':3.0/(N),
          'pole_offset':get_first_pole(gamma)}
    [p_values, cutts] = make_path(**args)
    if gammaNmax is not None and gamma*N > gammaNmax and gamma>gammaNmax:
        return {"K'splay":np.NaN, "K'twist":np.NaN, "K'bend":np.NaN}

    conditions = {}
    conditions["G000"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJG"] = np.zeros(len(p_values), dtype=complex)

    # Do stone fence calculations
    for ii, p in enumerate(p_values):
        G = calc_G_set(p, gamma, m_values=[0])["Gmll"]
        conditions["G000"][ii] = G[0][0,0]
        conditions["GJG"][ii] = GJG(G)

    # Do inverse laplace transforms
    data={}
    for key, values in conditions.items():
        reduction=(2.0*args["scale"]+args["pole_offset"])*N
        data[key]=np.real(invLaplace_path(p_values, cutts, values, N,
                                          reduction=reduction))
    
    data["aLAf"] = gamma*(N**2)*15*data["G000"]\
                   /(8*np.pi*data["GJG"])

    data["a2lpAf"] = data["aLAf"]/N
    return data

def get_Frank_values(gamma, N, gammaNmax=None, laplace_args={}):
    """Returns dictionary of Frank Elastic Values for specified system.

    Nondimensionalized Frank elastc constant K'=K*A/(phi_00*L).
    Maier saupe parameter ether nondimensionalized by L*A*phi_00 or
    2*l_p*A*phi_00 for aLAf and a2lpAf respectively.
    Correlation functions S are also returned.

    Args:
        gamma (real): Field strength in kT's per Kuhn length of chain
        N (real): How many Kuhn lengths long the polymers are
        gammaNmax (real): Cutoff to prevent wasting time.
        laplace_args (dict): Arguements to pass to inverse laplace calculation.


    Note: when N=0.0, gamma is assume to be gamma*N, that is kT's per chain.
    """
    if N==0.0:
        return RR_datapoint(gamma)
    args={'factor':400,
          'scale':3.0/(N),
          'pole_offset':get_first_pole(gamma)}
    for key, value in laplace_args.items():
        args[key]=value
        
    [p_values, cutts] = make_path(**args)
    if gammaNmax is not None and gamma*N > gammaNmax and gamma>gammaNmax:
        return {"K'splay":np.NaN, "K'twist":np.NaN, "K'bend":np.NaN}

    conditions = {}
    conditions["G000"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJG"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_00"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_11"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_couple"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_density"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_-1-1"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_22"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJG_-2-2"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJGJG_11_mk00"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJGJG_11_mk11"] = np.zeros(len(p_values), dtype=complex)
    conditions["GJGJGJG_-1-1_mk11"] = np.zeros(len(p_values), dtype=complex)

    # Do stone fence calculations
    for ii, p in enumerate(p_values):
        G = calc_G_set(p, gamma, m_values=[-3,-2,-1,0,1,2,3])["Gmll"]
        conditions["G000"][ii] = G[0][0,0]
        conditions["GJG"][ii] = GJG(G)
        conditions["GJGJG_00"][ii] = GJGJG(G, 2, 2, 0, 0)
        conditions["GJGJG_11"][ii] = GJGJG(G, 2, 2, 1, 1)
        conditions["GJGJG_couple"][ii] = GJGJG(G, 0, 2, 0, 0)
        conditions["GJGJG_density"][ii] = GJGJG(G, 0, 0, 0, 0)
        conditions["GJGJG_-1-1"][ii] = GJGJG(G, 2, 2, -1, -1)
        conditions["GJGJG_22"][ii] = GJGJG(G, 2, 2, 2, 2)
        conditions["GJGJG_-2-2"][ii] = GJGJG(G, 2, 2, -2, -2)
        conditions["GJGJGJG_11_mk00"][ii] = GJGJGJG(G, 2, 2, 1, 1, (0,0))
        conditions["GJGJGJG_11_mk11"][ii] = GJGJGJG(G, 2, 2, 1, 1, (1,1))
        conditions["GJGJGJG_-1-1_mk11"][ii] = GJGJGJG(G, 2, 2, -1, -1, (1,1))

    # Do inverse laplace transforms
    data={}
    for key, values in conditions.items():
        reduction=(2.0*args["scale"]+args["pole_offset"])*N
        data[key]=np.real(invLaplace_path(p_values, cutts, values, N,
                                          reduction=reduction))
    
    for (m1m2, mkmk) in [("11","11"), ("11","00"), ("-1-1", "11")]:
        data["ddk"+mkmk+" S'22"+m1m2]=4*(4*np.pi/5)*(2/(N**2))\
                                     *data["GJGJGJG_"+m1m2+"_mk"+mkmk]\
                                     /data["G000"]
    for m1m2 in ["11", "-1-1","00","22","-2-2"]:
        data["S'22"+m1m2] = (4*np.pi/5)*(2/(N**2))\
                             *data["GJGJG_"+m1m2]\
                             /data["G000"]
    data["S'0200"] = (4*np.pi/np.sqrt(5))*(2/(N**2))*data["GJGJG_couple"]\
                     /data["G000"]
    data["S'0000"] = (4*np.pi)*(2/(N**2))*data["GJGJG_density"]\
                     /data["G000"]

    data["phi_20/phi_00"] = (N**-1)*np.sqrt(4*np.pi/5.0)*\
                            data["GJG"]/data["G000"]

    #data["K'splay 2"] = 4*np.pi*(data["GJG"]**2)*data["GJGJGJG_11_mk11"]\
    #                  /((N*data["GJGJG_11"])**2 * data["G000"])
    #data["K'twist 2"] = 4*np.pi*(data["GJG"]**2)*data["GJGJGJG_-1-1_mk11"]\
    #                  /((N*data["GJGJG_-1-1"])**2 * data["G000"])
    #data["K'bend 2"] = 4*np.pi*(data["GJG"]**2)*data["GJGJGJG_11_mk00"]\
    #                  /((N*data["GJGJG_11"])**2 * data["G000"])

    data["K'splay"] = (2*np.pi/(N**2))*(data["phi_20/phi_00"]**2)*\
                        (data["S'2211"]**-2)*data["ddk11 S'2211"]
    data["K'twist"] = (2*np.pi/(N**2))*(data["phi_20/phi_00"]**2)*\
                        (data["S'22-1-1"]**-2)*data["ddk11 S'22-1-1"]
    data["K'bend"] = (2*np.pi/(N**2))*(data["phi_20/phi_00"]**2)*\
                        (data["S'2211"]**-2)*data["ddk00 S'2211"]

    data["aLAf"] = gamma*(N**2)*15*data["G000"]\
                   /(8*np.pi*data["GJG"])

    data["a2lpAf"] = data["aLAf"]/N

    #print("splay %f = %f"%(data["K'splay"],data["K'splay 2"]))
    #print("twist %f = %f"%(data["K'twist"],data["K'twist 2"]))
    #print("bend %f = %f"%(data["K'bend"],data["K'splay 2"]))
    return data


def test_Frank_accuracy(N=10, gamma=1000):
    """
    Compair result for different laplace_args.

    Looks really good at N=10, gamma=25
    To within ~0.3% for N=100, gamma=50. Increasing factor is impoartant.
    At N=100 gamma=100 the factor=400 level gets off by a few percent.
    At N=10, gamma=100 it is accurate to may descimal points
    At N=10, gamma=1000 there is a several persent error
    """
    
    args={'factor':400, 'maxp':6000, 'nwing':100}
    data = get_Frank_values(gamma, N, laplace_args=args)
    print(args)
    print([data["K'splay"], data["K'twist"], data["K'bend"]])

    args={'factor':400, 'maxp':24000, 'nwing':400}
    data = get_Frank_values(gamma, N, laplace_args=args)
    print(args)
    print([data["K'splay"], data["K'twist"], data["K'bend"]])

    args={'factor':1200, 'maxp':6000, 'nwing':100}
    data = get_Frank_values(gamma, N, laplace_args=args)
    print(args)
    print([data["K'splay"], data["K'twist"], data["K'bend"]])
#test_Frank_accuracy()





import numpy as np
from . import rouse


def generate_poisson_homologs(mu, chr_size=1):
    r"""
    Generate the number and location of linkage between homologous chromosomes

    Parameters
    ----------
    mu : float
        Average number of (uniformly distributed) linkages between chromosomes.
    chr_size : float
        Size of the chromosome.

    Returns
    -------
    cell : np.array of float
        List of linkage locations between the homologous chromosomes.

    """
    cell = np.sort(np.random.random_sample(size=np.random.poisson(lam=mu)))
    return chr_size*cell


def mscd(t, linkages, label_loc, chr_size, nuc_radius, b, D,
         num_modes=rouse._default_modes):
    r"""
    Calculate the MSCD for the model of linked chromosomes

    Parameters
    ----------
    t : float array
        Time in seconds
    linkages : float array
        List of the link positions between the homologous chromosomes
    label_loc : float
        Location of the fluorescent label along the chromosome (microns)
    chr_size : float
        Length of the chromosome (microns)
    nuc_radius : float
        Radius of the nucleus (microns)
    b : float
        Kuhn length (microns)
    D : float
        Diffusivity (microns ** 2 / second)
    num_modes : int
        Number of normal modes used in the calculation

    Returns
    -------
    mscd_model : float array (size len(t))
        Calculated MSCD (microns ** 2) for the model with defined linkages

    """
    linkages = np.array(linkages) / b
    chr_size = chr_size / b
    label_loc = label_loc / b

    # Evaluate the MSCD if there are no linkages
    if len(linkages) == 0:
        mscd_model = 2 * rouse.linear_mid_msd(t, b, chr_size, D, num_modes)
        return np.minimum(mscd_model, nuc_radius)

    # Evaluate the MSCD if there are linkages between the chromosomes
    i = np.searchsorted(linkages, label_loc)
    if i == 0:
        mscd_func = rouse.linear_mscd
        Ndel = linkages[0] - label_loc
        N = 2 * linkages[0]
    elif i == len(linkages):
        mscd_func = rouse.linear_mscd
        Ndel = label_loc - linkages[-1]
        N = 2 * (chr_size - linkages[-1])
    else:
        mscd_func = rouse.ring_mscd
        Ndel = linkages[i] - label_loc
        N = 2*(linkages[i] - linkages[i - 1])

    mscd_model = mscd_func(t, D=D, Ndel=Ndel, N=N, b=b, num_modes=num_modes)

    return np.minimum(nuc_radius, mscd_model)


def mscd_plateau(linkages, label_loc, chr_size, nuc_radius, b=1):
    r"""
    Evaluate the plateau values in the MSCD

    Parameters
    ----------
    linkages : float array
        List of the link positions between the homologous chromosomes
    label_loc : float
        Location of the fluorescent label along the chromosome (microns)
    chr_size : float
        Length of the chromosome (microns)
    nuc_radius : float
        Radius of the nucleus (microns)
    b : float
        Kuhn length (microns)

    Returns
    -------
    mscd_plateau : float
        Plateau value of the MSCD in the long-time asymptotic limit
        (microns ** 2).

    """
    chr_size /= b
    label_loc /= b
    linkages = np.array(linkages) / b

    # Evaluate the MSCD if there are no linkages
    if len(linkages) == 0:
        mscd_plateau = nuc_radius ** 2
        return mscd_plateau

    # Evaluate the MSCD if there are linkages between the chromosomes
    i = np.searchsorted(linkages, label_loc)
    if i == 0:
        Ndel = linkages[0] - label_loc
        mscd_plateau = 4 * b ** 2 * Ndel
    elif i == len(linkages):
        Ndel = label_loc - linkages[-1]
        mscd_plateau = 4 * b ** 2 * Ndel
    else:
        Ndel = linkages[i] - label_loc
        N = 2*(linkages[i] - linkages[i - 1])
        mscd_plateau = 2 * b ** 2 / (1 / (2 * Ndel) + 1 / (N - 2 * Ndel))

    return np.minimum(mscd_plateau, nuc_radius**2)



"""Fractional (Rouse) polymer. A Rouse polymer in a viscoelastic medium.

For the case alpha=1, these formulas should reduce to the analogous ones in
:mod:`wlcsim.analytical.rouse`."""
from .rouse import kp_over_kbt, rouse_mode, linear_mid_msd
from ..tabulation import frac_vel_corr

from bruno_util.mittag_leffler import ml as mittag_leffler
import numpy as np

from pathlib import Path

def frac_msd(t, alpha, D):
    """MSD of fractionally diffusing free particle.

    Weber, Phys Rev E, 2010 (Eq 10)"""
    return 3*D*np.sin(alpha*np.pi)*t**alpha \
        /(np.pi*(1-alpha/2)*(1-alpha)*alpha)

def frac_cv(t, alpha, D):
    """Velocity autocorrelation of a fractionally-diffusing particle.
    Weber, Phys Rev E, 2010 (Eq 32)"""
    return -(3*D)*np.sin(alpha*np.pi)/(np.pi*(2-alpha))*np.abs(np.power(t, alpha-2))

def frac_discrete_cv(t, delta, alpha, D):
    """Discrete velocity autocorrelation of a fractionally-diffusing particle.
    Weber, Phys Rev E, 2010 (Eq 33)"""
    t = np.atleast_1d(t)
    delta = np.atleast_1d(delta)
    # too many divide by t's to rewrite to avoid these warnings in less than
    # 5min, not worth it.
    with np.errstate(divide='ignore', invalid='ignore'):
        eta = delta/t
        t = t + np.zeros_like(eta) # fix to full size if not already
        delta = delta + np.zeros_like(eta) # fix to full size if not already
        cv_delta_t = frac_cv(t, alpha, D)/(eta*eta*alpha*(1 - alpha))
        cv_delta_t[eta<=1] = cv_delta_t[eta<=1] \
            *(2 - np.power(1 - eta[eta<=1], alpha) - np.power(1 + eta[eta<=1], alpha))
        cv_delta_t[eta>1] = cv_delta_t[eta>1] \
            *(2 + np.power(eta[eta>1] - 1, alpha) - np.power(1 + eta[eta>1], alpha)) \
            + frac_msd(delta[eta>1] - t[eta>1], alpha, D)/delta[eta>1]/delta[eta>1]
        cv_delta_t[t == 0] = frac_msd(delta[t == 0], alpha, D)/np.power(delta[t == 0], 2)
    return cv_delta_t

def frac_discrete_cv_normalized(t, delta, alpha):
    """Normalized discrete velocity autocorrelation of a fractionally-diffusing
    particle. Should be equivalent to

        frac_discrete_cv(t, delta, 1, 1)/frac_discrete_cv(0, delta, 1, 1)

    Lampo, BPJ, 2016 (Eq 5)"""
    return (np.power(np.abs(t - delta), alpha)
        - 2*np.power(np.abs(t), alpha)
        + np.power(np.abs(t + delta), alpha)
        )/(2*np.power(delta, alpha))

def terminal_relaxation_time(alpha, b, N, D):
    return np.power(N*N*b*b/D, 1/alpha)

def frac_rouse_mode_corr(p, t, alpha, b, N, D):
    """Weber Phys Rev E 2010, Eq. 21."""
    # equivalent to the following
    # kp = rouse_mode_coef(p, b, N, kbT)
    # return (3*kbT/kp)*mittag_leffler((-kp*t**alpha)/(N*xi*gamma(3-alpha)), alpha, 1)
    kp_norm = kp_over_kbt(p, b, N)
    # notice kp_over_kbt*D == kp/xi
    return 3*kp_norm*mittag_leffler(-(kp_norm*D)/(N*gamma(3-alpha)) * t**alpha, alpha, 1)

def frac_rouse_mid_msd(t, alpha, b, N, D, num_modes=1000):
    """
    Weber Phys Rev E 2010, Eq. 24.
    """
    if alpha == 1:
        return linear_mid_msd(t, b, N, D, num_modes)
    rouse_corr = 0
    for p in range(1, num_modes+1):
        # k2p = rouse_mode_coef(2*p, b, N, kbT)
        # rouse_corr += 12*kbT/k2p*(1 - mittag_leffler(-k2p*t**alpha/(N*xi*gamma(3-alpha)),
        #         alpha, 1))
        k2p_norm = kp_over_kbt(2*p, b, N)
        rouse_corr += (1/k2p_norm)*(1 - mittag_leffler(
                -(k2p_norm*D)/(N*gamma(3-alpha)) * t**alpha, alpha, 1
        ))
    # return rouse_corr + frac_msd(t, alpha, kbT, xi)/N
    return 12*rouse_corr + frac_msd(t, alpha, D)/N

def rouse_cv_mid(t, alpha, b, N, D, min_modes=1000):
    """Velocity autocorrelation of midpoint of a rouse polymer.

    Weber Phys Rev E 2010, Eq. 33."""
    t = np.atleast_1d(t)
    psum = np.zeros_like(t)
    for p in range(1, min_modes+1):
        # k2p = rouse_mode_coef(2*p, b, N, kbT)
        k2p_norm = kp_over_kbt(2*p, b, N)
        # coef = -k2p/(N*xi*gamma(3-alpha))
        coef = -(k2p_norm*D)/(N*gamma(3-alpha))
        psum += mittag_leffler(coef*np.power(t, alpha), alpha, alpha-1)
    gam = 1 # why Steph, why...
    return frac_cv(t, alpha, D)*(1 + 2*gam*(alpha - 1)*psum)

def tR(alpha, b, N, D):
    """Lampo et al, BPJ, 2016, eq 8"""
    return np.power(N*N*b*b/D, 1/alpha)

def tDeltaN(n1, n2, alpha, b, D):
    """Lampo et al, BPJ, 2016, eq 11"""
    delN = np.abs(n2 - n1)
    return np.power(delN*delN*b*b/D, 1/alpha)

def rouse_cvv_ep(t, delta, p, alpha, b, N, D):
    """Term in parenthesis in Lampo, BPJ, 2016 Eq. 10."""
    t = np.atleast_1d(t)
    delta = np.atleast_1d(delta)
    p = np.atleast_1d(p)
    tpdelta = np.power(np.abs(t + delta), alpha)
    tmdelta = np.power(np.abs(t - delta), alpha)
    ta = np.power(np.abs(t), alpha)
    # kp = rouse_mode_coef(p, b, N, kbT)
    kp_norm = kp_over_kbt(p, b, N)
    # z = -kp/(N*xi*gamma(3 - alpha))
    z = -(kp_norm*D)/(N*gamma(3-alpha))
    return -mittag_leffler(z*tpdelta, alpha, 1) \
        + 2*mittag_leffler(z*ta, alpha, 1) \
        - mittag_leffler(z*tmdelta, alpha, 1)


def frac_rouse_cvv(t, delta, n1, n2, alpha, b, N, D, min_modes=500,
              rtol=1e-5, atol=1e-8, force_convergence=True):
    """Velocity cross-correlation of two points on fractional Rouse polymer

    rtol/atol specify when to stop adding rouse modes. a particle (t,delta)
    pair is considered to have converged when np.isclose returns true give
    rtol/atol and p is even (p odd contributes almost nothing)

    Lampo, BPJ, 2016 Eq. 10."""
    t = np.atleast_1d(t)
    delta = np.atleast_1d(delta)
    tpdelta = np.power(np.abs(t + delta), alpha)
    tmdelta = np.power(np.abs(t - delta), alpha)
    ta = np.power(np.abs(t), alpha)
    # center of mass velocity correlation
    c0 = 3*D/(delta*delta*N*gamma(3-alpha)*gamma(1+alpha)) \
            * (tpdelta + tmdelta - 2*ta)
    # correction factor due to being on a polymer is an expansion in terms of
    # the normal modes of the polymer
    rouse_corr = np.zeros_like(tpdelta)
    # because the rouse modes are cos(p*pi*(n/N)), then the number of p's to
    # include per pass before checking for convergence again should be
    theta_eps = min(n1/N/2, n2/N/2)
    chunk_size = int(np.ceil(2*np.pi/theta_eps))
    # first include the first min_modes for all values of t
    for p in range(1, min_modes+1):
        # kp = rouse_mode_coef(p, b, N, kbT)
        kp_norm = kp_over_kbt(p, b, N)
        # z = -kp/(N*xi*gamma(3-alpha))
        z = -(kp_norm*D)/(N*gamma(3-alpha))
        pdiff = 3/kp_norm*rouse_mode(p, n1, N)*rouse_mode(p, n2, N) \
                *(2*mittag_leffler(z*ta, alpha, 1)
                - mittag_leffler(z*tpdelta, alpha, 1)
                - mittag_leffler(z*tmdelta, alpha, 1))
        rouse_corr = rouse_corr + pdiff
    counts = min_modes*np.ones_like(rouse_corr)
    if not force_convergence:
        return c0 + rouse_corr
    # now include more terms in chuck_size number of p's at a time until
    # convergence.
    p += 1
    old_rouse_corr = rouse_corr.copy() # force alloc
    tol_i = np.ones_like(rouse_corr).astype(bool) # always at least one correction term
    while np.any(tol_i):
        tpd = tpdelta[tol_i]
        tmd = tmdelta[tol_i]
        tai = ta[tol_i]
        old_rouse_corr[:] = rouse_corr # force memmove
        for i in range(chunk_size):
            # kp = rouse_mode_coef(p, b, N, kbT)
            kp_norm = kp_over_kbt(p, b, N)
            # z = -kp/(N*xi*gamma(3-alpha))
            z = -(kp_norm*D)/(N*gamma(3-alpha))
            pdiff = 3/kp_norm*rouse_mode(p, n1, N)*rouse_mode(p, n2, N) \
                    *(2*mittag_leffler(z*tai, alpha, 1)
                        - mittag_leffler(z*tpd, alpha, 1)
                        - mittag_leffler(z*tmd, alpha, 1))
            rouse_corr[tol_i] = rouse_corr[tol_i] + pdiff
            counts[tol_i] += 1
            p += 1
        tol_i = ~np.isclose(rouse_corr, old_rouse_corr, rtol, atol)
    return c0 + rouse_corr, counts

def rouse_nondim_(t, delta, n1, n2, alpha, b, N, D):
    """uses parameters defined by Lampo et al, BPJ, 2016, eq 12"""
    tOverDelta = t/delta
    deltaOverTDeltaN = delta/tR(alpha, b, N, D) \
            *np.power(np.abs(n2-n1)/N, -2/alpha)
    return tOverDelta, deltaOverTDeltaN, alpha
rouse_nondim = np.vectorize(rouse_nondim_)

def un_rouse_nondim(tOverDelta, deltaOverTDeltaN, alpha, delN=0.001):
    """Takes a requested choice of parameters to compare to Tom's calculated
    values and generates a full parameter list that satisfies those
    requirements.

    uses approximation defined by Lampo et al, BPJ, 2016, eq 12

    In Tom's paper, delN is non-dimensinoalized away into deltaOverTDeltaN, but
    that only works in the case where the chain is infinitely long. Otherwise,
    where exactly we choose n1,n2 will affect the velocity correlation because
    of the effects of the chain ends. While other choices are truly arbitrary
    (e.g. D, b, N, etc), delN can be used to specify the size of the segment
    (relative to the whole chain) used to compute the cross correlation."""
    # get arbitrary choices out of the way first
    N = 1
    b = 1
    D = 1
    n1 = N/2
    n2 = N/2 + delN
    # now the computed quantities
    tr = tR(alpha, b, N, D)
    delta = deltaOverTDeltaN*tr*np.power(delN/N, 2/alpha)
    t = tOverDelta*delta
    return t, delta, n1, n2, alpha, b, N, D

def vc(t, delta, beta):
    """velocity correlation of locus on rouse polymer. beta = alpha/2."""
    return ( np.power(np.abs(t - delta), beta)
           + np.power(np.abs(t + delta), beta)
           - 2*np.power(np.abs(t), beta)
           )/( 2*np.power(delta, beta) )

def vcp(t, delta, beta):
    return ( np.power(np.abs(t + delta), beta)
           - np.power(np.abs(t), beta) )*np.power(delta, beta)

def vvc_unscaled_theory(t, delta, beta, A, tDeltaN):
    """velocity cross correlation of two points on rouse polymer."""
    return 2*(vc(t*delta, delta, beta) - frac_vel_corr(t, delta/tDeltaN, 2*beta))

def vvc_rescaled_theory(t, delta, beta, A, tDeltaN):
    """velocity cross correlation of two points on rouse polymer."""
    return 2*A*np.power(delta, beta)*(vc(t*delta, delta, beta) - frac_vel_corr(t, delta/tDeltaN, 2*beta))

def vvc_normalized_theory(t, delta, beta, A, tDeltaN):
    return vvc_unscaled_theory(t, delta, beta, A, tDeltaN) \
         / vvc_unscaled_theory(0, delta, beta, A, tDeltaN)




"""For now, just a widget that requires hardcoding everythign about hte
simulation. Eventually, a tool to load in simulation data from pysim or
wlcsim and vizualize it intelligently."""
import numpy as np
import matplotlib
matplotlib.use("Qt4Agg") # This program works with Qt only
from matplotlib.widgets import Slider, Button, RadioButtons
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt
import os
import re
from .data import Sim


class SimulationViewer(object):

    def __init__(self, x=None):
        # will keep track of which *index* in the time axis should be used to
        # draw the current scene
        self.t = 0
        self.sim = None
        # define the actual axes layout
        self.fig = plt.figure()
        # area to plot polymers
        self.ax3d = plt.axes([0.05, 0.15, 0.9, 0.8], facecolor='w', projection='3d')
        # area to plot "slider" for selecting what time to plot
        self.ax = plt.axes([0.1, 0.025, 0.8, 0.05], facecolor='lightgoldenrodyellow')
        # set up the QtGui panel and textbox
        self.root = self.fig.canvas.manager.window
        self.panel = QtWidgets.QWidget()
        self.hbox = QtWidgets.QHBoxLayout(self.panel)
        self.textbox = QtWidgets.QLineEdit(parent=self.panel)
        self.hbox.addWidget(self.textbox)
        self.panel.setLayout(self.hbox)
        self.dock = QtWidgets.QDockWidget("Simulation Directory", self.root)
        self.root.addDockWidget(Qt.BottomDockWidgetArea, self.dock)
        self.dock.setWidget(self.panel)
        # handler to check if new simulation directory has been entered
        self.textbox.textChanged.connect(self.update_dir)
        # slider to control what "time" is plotted, (0,1) is rescaled total
        # simulation time units
        self.slider_t = Slider(self.ax, 't', 0, 1, valinit=0)
        # handler to check if the "time" slider has been moved
        self.slider_t.on_changed(self.update_t)
        plt.show()

    # should never be called before a Sim has been loaded successfully
    def update_drawing(self):
        num_polymers = self.sim.num_polymers
        if num_polymers != len(self.ax3d.lines):
            self.ax3d.lines = []
            for i in range(num_polymers):
                self.ax3d.plot(self.sim.r[self.t,0,i,:],
                               self.sim.r[self.t,1,i,:],
                               self.sim.r[self.t,2,i,:])
        for i,line in enumerate(self.ax3d.lines):
            x, y, z = (self.sim.r[self.t,0,i,:], self.sim.r[self.t,1,i,:],
                    self.sim.r[self.t,2,i,:])
            line.set_data(x, y)
            line.set_3d_properties(z)
        self.fig.canvas.draw_idle()

    def update_ax_limits(self):
        self.ax3d.set_xlim((np.nanmin(self.sim.r[:,0,:,:]), np.nanmax(self.sim.r[:,0,:,:])))
        self.ax3d.set_ylim((np.nanmin(self.sim.r[:,1,:,:]), np.nanmax(self.sim.r[:,1,:,:])))
        self.ax3d.set_zlim((np.nanmin(self.sim.r[:,2,:,:]), np.nanmax(self.sim.r[:,2,:,:])))

    def update_dir(self):
        text = self.textbox.text()
        try:
            self.sim = Sim(text)
        except FileNotFoundError:
            return
        self.update_t(self.slider_t.val)
        self.update_ax_limits()
        self.update_drawing()

    def update_t(self, val):
        if self.sim is None:
            return
        # calculate value on slider
        new_t = int(round(val*self.sim.num_time_points))
        # bound it to 0, num_time_points-1
        new_t = max(0, min(self.sim.num_time_points-1, new_t))
        self.t = new_t
        self.update_drawing()

if __name__ == '__main__':
    sim_viewer = SimulationViewer()




"""lines_in_box monte carlo simulator
pseudocode for col checker
poca - two points at which infinite lines get closest
bpoca - poca of two finite segments defining centers of cylinders
loca - line conencts poca
for periodization old line
  for periodization of new line
      if d(poca) > 2r
          return false
      if bpoca is not on tip of cylinder for either cylinder
          return naive formula
      elif one tip and one interior point (of closest approach)
          if loca's midpoint is inside box # already know: |loca| < 2 r
                  or check if intersections with box bdry collide
              return true
          else:
              return false
      else (if two tips)
          for each tip
              for each edge adjacent to the face tip is hitting
                  if |loca btween edge and cylinder| < r
                     mark wall opposite that edge for that tip
          for each wall that both tips share
              return if their ellipse outlines on that wall collide inside of the box wall

Yes I know this could have been done more prettily with object oriented shapes,
but I'm busy getting real results, do it yourself if you want.

"""
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
import pandas as pd
from ..utils import path as wpath
import os
from numba import jit
import multiprocessing
import pscan
from scipy.optimize import curve_fit
import statsmodels.api as sm
from scipy.spatial import cKDTree


# our "floating precision" cutoff
SMALL_NUM = np.power(10.0, -8)

# in what follows, a cylinder's centerline's is used as a proxy for the
# cylinder itself. comments often only make sense with this in mind (e.g. the
# cylinder "hits" a wall means the intersection of its centerline with the
# plane defining the wall).

# if a cylinder hits one of the infinite planes defined by the ws, w0s below at
# a distance far enough away from the unit box's extents, then the outline of
# the intersection of the full cylinder with the unit box's wall will be
# effectively a rectange instead of an ellipse. this is the maximum magnitude
# of the collision (in units of box-widths away from the box's center) before the
# collision code simply uses the parallelogram approximation
max_boxes_away_for_ellipse_collision = 20

# now comes an absurd amount of precomputed constants that describe a unit box
# in the first quadrant

# normal vectors and intercepts defining the planes of the unit box in the
# first quadrant, used throughout collision detection
ws = [[0.0, -1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0],
      [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]
ws = [np.array(w) for w in ws]
# three planes pass through (0,0,0), other three pass through (1,1,1)
w0s = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
# how to project a vector into any of the box walls
plane_projs = [np.array(w) == 0.0 for w in ws]
non_zero_indices = [[i for i,c in enumerate(plane_proj) if c]
                    for plane_proj in plane_projs]
plane_projs = [p.astype(float) for p in plane_projs]
# convenience variables for defining points on the box
ux = np.array([1.0, 0.0, 0.0])
uy = np.array([0.0, 1.0, 0.0])
uz = np.array([0.0, 0.0, 1.0])
uo = np.array([0.0, 0.0, 0.0])
edges = [(uo, ux), (ux, ux+uz), (ux+uz, uz), (uz, uo),
         (uo, uy), (uy, uy+uz), (uy+uz, uz),
         (ux, ux+uy), (ux+uy, uy),
         (ux+uy, ux+uy+uz), (ux+uz, ux+uy+uz), (uy+uz, ux+uy+uz)]
# I didn't choose the best edge labelling, but if you want to change it, just
# change all the relevant variables up here in the global scope, I didn't use
# the edge labeling explicitly anywhere in the main program, just through these
# variables

# each edge is adjacent to two faces
adjacent_faces = [(0, 2), (0, 4), (0, 5), (0, 1),
                  (1, 2), (1, 3), (1, 5), (2, 4),
                  (2, 3), (3, 4), (4, 5), (3, 5)]
# a map from faces to clockwise (from inside box) listing of adjacent edges to
# that face, list of lists of edge indexes.
adjacent_edges = [[0, 1, 2, 3],  # face 0
                  [4, 5, 6, 3],  # face 1
                  [0, 4, 8, 7],  # face 2
                  [11, 5, 8, 9], # face 3
                  [1, 10, 9, 7], # face 4
                  [2, 6, 11, 10]] # face 5
# face, edge -> other face
face_edge_to_face = {(faces[0], edge): faces[1] for edge, faces
        in enumerate(adjacent_faces)}
face_edge_to_face.update({(faces[1], edge): faces[0] for edge, faces
        in enumerate(adjacent_faces)})
# set of pairs of faces that share an edge, with both index orders for fast
# lookup
face_pairs = {face_pair for face_pair in adjacent_faces}
face_pairs.update({(face_pair[1], face_pair[0]) for face_pair in adjacent_faces})
# lookup the opposite of a face quickly as face_opposite[face]
face_opposite = [3, 4, 5, 0, 1, 2]
# periodization vectors to add to get possible periodizations about a
# particular face can be looked up easily here
# we quickly use the normals of the faces that share an edge with the face we
# want to keep constant during our periodization
periodizers_given_face = [
        [ws[face_edge_to_face[(face,edge)]] for edge in adjacent_edges[face]]
        for face in range(6)]
for plist in periodizers_given_face:
    plist.append(plist[0] + plist[1])
    plist.append(plist[1] + plist[2])
    plist.append(plist[2] + plist[3])
    plist.append(plist[3] + plist[0])

# semantically useful throughout code, create once
box_center = np.array([0.5, 0.5, 0.5])

def new_object(shape_params, shape):
    if 'sphere' in shape:
        r = shape_params
        return (uniform_point_from_unit_cube(), r)
    elif 'cylinder' in shape:
        return uniform_segment_from_unit_cube()
    else:
        raise ValueError('Invalid shape requested!')
    return None

def volume_of_shape(obj, shape_params, shape, periodic):
    if 'sphere' in shape:
        (c, r) = obj
        return volume_of_sphere(c, r, periodic)
    elif 'cylinder' in shape:
        xi, xf = obj
        d = shape_params
        return volume_of_cylinder(xi, xf, d, periodic)
    else:
        raise ValueError('Invalid shape requested!')
    return np.nan

def detect_collision(obj, obj_new, shape_params, shape, periodic=False):
    if 'sphere' in shape:
        (c, r) = obj
        (c_new, r_new) = obj_new
        return detect_collision_spheres(c, r, c_new, r_new, periodic)
    elif 'cylinder' in shape:
        xi, xf = obj
        xi_new, xf_new = obj_new
        d = shape_params
        return detect_collision_cylinders(xi, xf, xi_new, xf_new, d, periodic)
    else:
        raise ValueError('Invalid shape requested!')
    return False

def columns_for_shape(shape):
    if 'cylinder' in shape:
        return ['xi1', 'xi2', 'xi3', 'xf1', 'xf2', 'xf3', 'phi', 'num_objects',
                'did_succeed', 'is_removal']
    elif 'sphere' in shape:
        return ['c1', 'c2', 'c3', 'r', 'phi', 'num_objects',
                'did_succeed', 'is_removal']
    else:
        raise ValueError('Invalid shape requested!')
    return []

def numeric_columns_for_shape(shape):
    if 'sphere' in shape:
        return ['c1', 'c2', 'c3', 'r', 'phi', 'num_objects']
    elif 'cylinder' in shape:
        return ['xi1', 'xi2', 'xi3', 'xf1', 'xf2', 'xf3', 'phi', 'num_objects']
    else:
        raise ValueError('Invalid shape requested!')
    return []

def row_for_shape(obj, shape, phi, num_objects, did_succeed, is_removal):
    if 'sphere' in shape:
        if obj is None:
            c = [np.nan, np.nan, np.nan]
            r = np.nan
        else:
            c, r = obj
        return [c[0], c[1], c[2], r, phi, num_objects, did_succeed, is_removal]
    elif 'cylinder' in shape:
        if obj is None:
            xf = [np.nan, np.nan, np.nan]
            xi = [np.nan, np.nan, np.nan]
        else:
            xi, xf = obj
        return [xi[0], xi[1], xi[2], xf[0], xf[1], xf[2], phi, num_objects,
                did_succeed, is_removal]
    else:
        raise ValueError('Invalid shape requested!')
    return []

def get_removal_amount(obj, shape):
    """Get parameter that is used to scale probability of removal.
    e.g. for cylinders, this is their length. for spheres, this is their
    radius. etc."""
    if 'sphere' in shape:
        return 1 # all have same volume, so do per number
    elif 'cylinder' in shape:
        xi, xf = obj
        return np.linalg.norm(xf - xi)
    else:
        raise ValueError('Invalid shape requested!')
    return np.nan

@jit(nopython=True)
def line_hit_hyperplane(xi, xf, w, w0):
    """Get point of intersection of a lines in N-D with a hyperplane of
    dimension N-1. The line should be defined by two pionts and the hyperplane
    should be defined by its normal vector and a point on it."""
    w = w/np.linalg.norm(w)
    x = xf - xi
    # projection onto plane's normal of x
    x_proj_w = np.dot(w, x)
    if np.abs(x_proj_w) < SMALL_NUM: # effectively parallel to plane
        return None
    # intersection point as parameter of xi + (xf - xi)*t
    t = (w0 - np.dot(xi, w))/x_proj_w
    # intersection point itself
    return xi + x*t

@jit(nopython=True)
def get_face_given_point(x):
    # xz-plane
    if np.abs(x[1] - 0.0) < SMALL_NUM:
        return 0
    # yz-plane
    elif np.abs(x[0] - 0.0) < SMALL_NUM:
        return 1
    elif np.abs(x[2] - 0.0) < SMALL_NUM:
        return 2
    elif np.abs(x[1] - 1.0) < SMALL_NUM:
        return 3
    elif np.abs(x[0] - 1.0) < SMALL_NUM:
        return 4
    elif np.abs(x[2] - 1.0) < SMALL_NUM:
        return 5

@jit(nopython=True)
def d_line_line(x1, x2, y1, y2):
    """Find distance between two lines defined by two non-degenerate
    points on each of those lines."""
    w = x1 - y1
    u = x2 - x1
    v = y2 - y1
    uu = np.dot(u,u)
    uv = np.dot(u,v)
    vv = np.dot(v,v)
    uw = np.dot(u,w)
    vw = np.dot(v,w)
    D = uu*vv - uv*uv # "descrimant" / denominator ( always > 0 )

    if D < SMALL_NUM: # lines must be nearly parallel
        # subtract projection of w onto either line from w
        # same as distance formula from just one line to a point
        s = 0.0
        t = uw/uu if uu > vv else vw/vv
    else:
        s = (uv*vw - vv*uw)/D
        t = (uu*vw - uv*uw)/D
    # vector of closest approach
    voca = w + s*u - t*v
    return np.linalg.norm(voca)


@jit(nopython=True)
def d_segment_segment(xinit0, xfinal0, xinit1, xfinal1):
    """Find distance between two segments defined by initial
    and ending position vectors."""
    # see e.g.
    # http://geomalgorithms.com/a07-_distance.html#dist3D_Segment_to_Segment()
    # for details of algorithm
    u = xfinal0 - xinit0
    v = xfinal1 - xinit1
    w = xinit0 - xinit1
    a = np.dot(u,u)
    b = np.dot(u,v)
    c = np.dot(v,v)
    d = np.dot(u,w)
    e = np.dot(v,w)
    D = a*c - b*b
    sc = D; sN = D; sD = D;
    tc = D; tN = D; tD = D;

    if D < SMALL_NUM:
        sN = 0.0         # force using point P0 on segment S1
        sD = 1.0         # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:                 # get the closest points on the infinite lines
        sN = (b*e - c*d)
        tN = (a*e - b*d)
        if sN < 0.0:        # sc < 0 => the s=0 edge is visible
            sN = 0.0
            tN = e
            tD = c
        elif sN > sD:  # sc > 1  => the s=1 edge is visible
            sN = sD
            tN = e + b
            tD = c
    if tN < 0.0:            # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        if -d < 0.0:
            sN = 0.0
        elif -d > a:
            sN = sD
        else:
            sN = -d
            sD = a
    elif tN > tD:      # tc > 1  => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        if (-d + b) < 0.0:
            sN = 0
        elif (-d + b) > a:
            sN = sD
        else:
            sN = (-d +  b)
            sD = a
    # finally do the division to get sc and tc
    sc = sN / sD if np.abs(sN) > SMALL_NUM else 0.0
    tc = tN / tD if np.abs(tN) > SMALL_NUM else 0.0

    # get the difference of the two closest points
    dP = w + (sc * u) - (tc * v)  # =  S1(sc) - S2(tc)

    return np.linalg.norm(dP)

@jit(nopython=True)
def norm3_squared(x):
    """Avoid extra operations when doing something like np.dot(f(x), f(x))"""
    # we don't need the generality of np.linalg.norm, so @jit of this simple function
    # will probably outperform np
    return x[0]*x[0] + x[1]+x[1] + x[2]*x[2]
    # return (x[0] - y[0])**2 + (x[1] - y[1])**2 + (x[2] - y[2])**2

def periodic_cylinder_collision(ui, uf, vi, vf, d):
    """Decide whether or not two cylinders inside of the box of side length 1
    are overlapping each other, when the box has "periodic" boundary conditions
    along all sides except for those on which the cylinders terminate."""

#     # renormalize to the equivalent problem on unit cube
#     if a != 1:
#         d = d/float(a)
#         ui = ui/a
#         vi = vi/a
#         uf = uf/a
#         vf = vf/a
#         a = 1.0
    r = d/2.0
    u = uf - ui
    v = vf - vi
    a = np.dot(u,u)
    b = np.dot(u,v)
    c = np.dot(v,v)
    d = np.dot(u,w)
    e = np.dot(v,w)
    D = a*c - b*b

    # if lines are nearly parallel, don't try to find point they're closest
    if D < SMALL_NUM:
        # any point works to check for collisions
        return norm3_squared(ui - vi) < d*d
    # calculate fractional path from xinit to xfinal that point of closest
    # approach between the two infinite lines defined by the input points is
    sN = (b*e - c*d) # numerator of "sC" on geomalgorithms.com
    tN = (a*e - b*d) # numerator of "tC"
    # if that point of closest approach is within the box (i.e. 0-1.0
    # fractionally), then collision detection is straightforward
    if 0.0 < sN and sN < D and 0.0 < tN and tN < D:
        sC = sN / sD if np.abs(sN) > SMALL_NUM else 0.0
        tC = tN / tD if np.abs(tN) > SMALL_NUM else 0.0
        return norm3_squared(ui + (sC*u) - (vi + (tC*v))) < d*d
    # if the point of closest approach is described by the tip of one
    # cylinder's center and an interior point of the other cylinder's center,
    # then there is a collision IFF (midpoint of line joining closest approach
    # points is inside of the box OR their extents on the box's wall overlap)
    # we take care of the first case here
    elif (sN <= 0.0 or D <= sN) and (0.0 < tN and tN < D):
        # uC = ui + (sC*u)
        # vC = vi + (tC*v)
        # wc = uC - vC
        # mp = vC + wc/2
        mp = ((ui + sC*u) + (vi + tC*v))/2
        # mp in unit box in first quadrant is same as ||R(pi/2)*(mp-0.5)||_1 < 1
        # but not sure if a rotation matrix is faster or slower than a bunch of
        # if statements
        if np.all((0.0 < mp) & (mp < 1.0)):
            return True
    # at this point, we know the point of closest approach is described by
    # either one tip and one interior point, and that we need to check if their
    # extents on the box's wall overlap, or the stick's two tips describe their
    # pointof closest approach, in which case we still only need to check if
    # their extents on the box overlap

    # then the sticks come closer and closer until they both hit their
    # respective ending walls. so we need only figure out what these
    # walls are and check if their outlines on the wall overlap or
    # not (accounting for fact that they might be fat enough for their
    # outlines to "spill" over onto a neighboring wall).
    # we number the walls 0-5, as in uniform_segment_from_unit_cube xz,
    # yz, xy planes and their opposites in the same order
    ui_walls = get_walls_with_outline(ui) #TODO periodize
    uf_walls = get_walls_with_outline(uf)
    vi_walls = get_walls_with_outline(vi)
    vi_walls = get_walls_with_outline(vf)
    isCollided = False
    for i in range(6):
        # if they cylinders' outlines on the box both have segment on a
        # particular wall, and these outlines overlap on the box's
        # wall, then the cylinders themselves overlap
        if (ui_walls[i] or uf_walls[i]) and (vi_walls[i] or vf_walls[i]) \
                and check_outline_overlap(ui, uf, vi, vf, wall=i):
            isCollided = True
            break
    return isCollided

def check_outline_overlap(ui, uf, vi, vf, wall):
    ellipse1 = get_ellipse_from_cylinder_wall(ui, uf, wall)
    ellipse2 = get_ellipse_from_cylinder_wall(vi, vf, wall)
    if ellipse1 and ellipse2:
        return check_ellipse_collisions(ellipse1, ellipse2)
    elif ellipse1:
        tube2 = get_tube2_from_cyclinder_wall(vi, vf, wall)
        return check_ellipse_tube2_collisions(ellipse=ellipse1, tube=tube2)
    elif ellipse2:
        tube1 = get_tube2_from_cyclinder_wall(ui, uf, wall)
        return check_ellipse_tube2_collisions(ellipse=ellipse2, tube=tube1)
    else:
        tube1 = get_tube2_from_cyclinder_wall(ui, uf, wall)
        tube2 = get_tube2_from_cyclinder_wall(vi, vf, wall)
        return check_tube2_tube2_collisions(tube1, tube2)

def get_ellipse_from_cylinder_wall(xi, xf, wall, r):
    w = ws[wall]
    w0 = w0s[wall]
    non_zero_index = non_zero_indices[wall]
    x = xf - xi
    Ip = line_hit_hyperplane(xi, xf, w, w0)
    # maybe they're not effectively parallell by numerical precision
    # (SMALL_NUM) test, but still effectively parallel in practice...say
    # collision happens very far from actual box
    if Ip is None \
    or norm3_squared(Ip - box_center) > MAX_BOXES_AWAY_FOR_ELLIPSE_COLLISION**2:
        return None
    # otherwise, we've determined the coordinates in the plane of interest
    h = Ip[non_zero_index[0]]
    k = Ip[non_zero_index[1]] # other index of Ip should be 1 or 0
    # angle from first coordinate to direction of long axis of ellipse
    theta = np.arctan(x[non_zero_index[1]]/x[non_zero_index[0]])
    # the minor axis is always the radius of the cylinder by definition
    b = r
    # the major axis is r*sec(phi), where phi is the (positive) angle between x and w
    # but we know phi = arccos(w.(x/||x||)), and sec(arccos(z)) = 1/z, so
    a = 1.0/np.dot(w, x/np.linalg.norm(x))
    return (a, b, theta, h, k)

def get_tube2_from_cyclinder_wall(xi, xf, wall, r):
    """Gets extents of 2D tube in which a (plane) wall, cylinder collision is
    effectively taking place for cylinders that are nearly perpendicular to the
    normal vector of the plane. This is returns as a start an end vector and a
    diameter in the plane."""
    w = ws[wall]
    w0 = w0s[wall]
    non_zero_index = non_zero_indices[wall]
    # the vector is nearly parallel to the plane, so we need only get the
    # distance from the plane to any point on the cylinder's center line to
    # determine its distance from the wall
    # also, w is already normalized in its definition
    dist = np.abs(d + np.dot(w, xi))
    # a plane parallel to a cylinder with radius r has a rectangular
    # intersection with the plane in the coordinates of the cylinder. the
    # height of this rectangle is just
    h = 2*np.sqrt(r*r - dist*dist)
    pi = xi*non_zero_index
    pf = xf*non_zero_index
    return (pi, pf, h)

def check_tube2_tube2_collisions(tube1, tube2):
    """Checks if the rectangular croos-sections of two cylinders parallel to
    the box's wall intersect in the the plane of the wall. The 2D tubular
    cross sections are passed in via two points along their center line and the
    the size of their extent as a distance from this center line. Thus, we need
    only check if the two corresponding infinite tube collide, then if they
    collide within the box's wall's boundaries. Two infinite tubes in 2D always
    collide unless their center lines are parallel. When they collide, their
    overlapping volume is a parallelogram. Either this parallelogram is
    contained by the box wall, or the two intersect. If they intersect, since
    they are both convex hulls of four points, it is enough to check that none
    of the vertices of the parallelgoram are contained in the box (and
    vice-versa) to conclude that they are not colliding."""
    xi, xf, h1 = tube1
    yi, yf, h2 = tube2
    x = xf - xi
    ui = x/np.linalg.norm(x)
    y = yf - yi
    vi = y/np.linalg.norm(y)
    # first check if tubes are parallell. if so, just return collision by
    # checking if the two lines as a whole are the correct distance from each
    # other
    if np.linalg.norm(ui - vi) < SMALL_NUM or np.linalg.norm(ui + vi) < SMALL_NUM:
        # distance between two parallel lines each defined by two points isn't
        # too hard to work out, just take the vector between any two points on
        # the line, then subtract off that vector's projection onto either line
        return np.linalg.norm((yi - xi) - np.dot(yi - xi, vi)) < h1 + h2
    # get the four vertices of the parallelogram that the two
    # tubes intersect in
    raise NotImplementedError('not done coding check_tube2_tube2_collisions')


def check_ellipse_collisions(ellipse1, ellipse2):
    a1,b1,theta1,h1,k1 = ellipse1
    a2,b2,theta2,h2,k2 = ellipse2
    raise NotImplementedError('not done coding check_ellipse_collisions')

@jit(nopython=True)
def uniform_point_from_unit_cube():
    return np.random.rand(3)

@jit(nopython=True)
def uniform_segment_from_unit_cube():
    """
    TODO: check if drawing a line through a random interior point with a random
    angle gives different distribution of stick density in box.
    Draws uniformly random segments from (unit) cube (in first quadrant)"""
    # sample a unit vector perpendicular to the xy-plane
    # using the "three normals" trick to generate a point on a sphere, then
    # ignoring the sign of the third component
    face_dx = np.random.randn(3)
    face_dx = face_dx/np.linalg.norm(face_dx)
    face_dx[2] = np.abs(face_dx[2]) # ignore sign
    # draw one of the six faces of the cube
    face = int(np.floor(np.random.rand(1)*6)[0]) # awkward casting to help jit
    # sample a point uniformly on that face
    face_x = np.random.rand(2)
    xinit = np.zeros(3)
    dxinit = np.zeros(3)
    # transform xinit and face_dx into face's coordinates
    if face == 0: # in xz-plane
        xinit[0] = face_x[0]
        xinit[2] = face_x[1]
        dxinit[1] = face_dx[2]
        dxinit[0] = face_dx[0]
        dxinit[2] = face_dx[1]
    elif face == 1: # in yz-plane
        xinit[1] = face_x[0]
        xinit[2] = face_x[1]
        dxinit[0] = face_dx[2]
        dxinit[1] = face_dx[0]
        dxinit[2] = face_dx[1]
    elif face == 2: # in xy-plane
        xinit[0:2] = face_x
        dxinit = face_dx
    elif face == 3: # opposite xz-plane
        xinit[0] = face_x[0]
        xinit[1] = 1.0
        xinit[2] = face_x[1]
        dxinit[1] = -face_dx[2]
        dxinit[0] = face_dx[0]
        dxinit[2] = face_dx[1]
    elif face == 4: # opposite yz-plane
        xinit[0] = 1.0
        xinit[1] = face_x[0]
        xinit[2] = face_x[1]
        dxinit[0] = -face_dx[2]
        dxinit[1] = face_dx[0]
        dxinit[2] = face_dx[1]
    elif face == 5: # opposite xy-plane
        xinit[0:2] = face_x
        xinit[2] = 1.0
        dxinit = face_dx
        dxinit[2] = -dxinit[2]
    #DEBUG
    # else:
    #     raise ValueError('internal error: face \\notin {0,1,2,3,4,5}')

    # now get collision point with other side of cube, by pretending our
    # segment is a path and asking at what time the path intersects the cube
    # if dx_i < 0, then x_i + dx_i*t exits the cube when x_i == 0
    # if dx_i > 0, then at x_i == 1
    tout = -xinit/dxinit # waste a couple cycles
    is_dx_pos = dxinit > 0
    if np.any(is_dx_pos):
        tout[is_dx_pos] = np.min((1.0 - xinit[is_dx_pos])/dxinit[is_dx_pos])
    xfinal = xinit + np.min(tout)*dxinit
    #DEBUG
    # if np.any(xfinal > 1.0):
    #     raise ValueError('internal error: xfinal should be inside unit cube')

    # get the segment initial and final locations
    return xinit, xfinal

@jit(nopython=True)
def uniform_segment_from_cube(a):
    """Draws uneformly random segments from cube (in first quadrant) defined by
    its side length."""
    xi,xf = uniform_segment_from_unit_cube()
    return a*xi, a*xf

@jit(nopython=True)
def volume_of_sphere(c, r, periodic=True):
    if periodic:
        return 4.0/3.0*np.pi*r*r*r
    else:
        raise NotImplementedError('volume of non-periodic sphere')
        return np.nan

# @jit(nopython=True)
def volume_of_cylinder(xi, xf, d, periodic=True):
    """Calculat the volume of a cylinder with center running from xi to xf
    inside a cube of side length a in the 1st quadrant. The formula used is the
    naive one, \pi r^2 l, but it is exactly correct if you're using periodic
    boundary conditions."""
    #TODO: ultra-low priority: implement exact formula for case when we are not
    # using periodic BCs
    # incorrect method: (see notes for full, correct method)
    # cases:
    #     1) hitting no edges - naive calculation works, same "in" as out
    #     2) hitting one edge - small modification
    #     3) hitting two+edges - ignored for now, hope these events are "rare"
    # case 1) d^2*pi*L is the correct formula. draw the "extra" volume you'd be
    # adding and subtracting from the "corners" that "jut out" of the box to
    # convince yourselves that they cancel out and you get back the naive
    # formula.
    # case 2) if you drew the first case out, you saw that the intersection of
    # the cylinder with the box edge is an ellipse, with major and minor axes
    # determined by the angle of intersection as (r|sec(\theta)|, r),
    # respectfully, where theta is angle from the normal to the wall to the
    # normal of the cylinder's natural cap. thus, you can consider case two as
    # the same as case 1, but subtracting the part ommited by the "collision"
    # with the wall not intersected by the center of the cylinder. since
    # finding the equation for this volume to be omitted seems hard in general,
    # since the ellipse's axes will in general not line up with the box edge,
    # we approximate by an intersection that is instead a circle with the
    # average of the major and minor axes as
    if periodic:
        xi, xf = get_true_xif_periodic(xi, xf)
    return np.pi*d*d*np.linalg.norm(xf - xi)

# @jit(nopython=True)
def get_true_xif_periodic(xi, xf):
    xi_face = get_face_given_point(xi)
    xf_face = get_face_given_point(xf)
    # if the faces are opposite of each other, then we are already fine
    if (xi_face, xf_face) not in face_pairs:
        return (xi, xf)
    # otherwise, we have to calculate which one is at a less extreme angle and
    # extend the other point beyond the box (recall ws are already normalized)
    mag_x = np.linalg.norm(xf - xi)
    thetai = np.arccos(np.dot(xi - xf, ws[xi_face])/mag_x)
    thetaf = np.arccos(np.dot(xf - xi, ws[xf_face])/mag_x)
    # if the angle is smaller, then the vector is more parallel to the normal
    # to the face, so we want to accept that face as the one we won't periodize
    # about
    if thetai > thetaf:
        # then we want new xi, found by continuing xf + t*(xi - xf) until it
        # hits the plane opposite to xf_face
        opposite_face = face_opposite[xf_face]
    else:
        opposite_face = face_opposite[xi_face]
    w = ws[opposite_face]
    w0 = w0s[opposite_face]
    # intersection point must always exist since we chose the plane of
    # intersection cleverly. order of xi,xf don't matter, the point will be
    # found regardless
    Ip = line_hit_hyperplane(xi, xf, w, w0)
    if thetai > thetaf:
        return (Ip, xf)
    else:
        return (xi, Ip)

def calculate_sphericity(lines):
    vectors = [line[1] - line[0] for line in lines]
    # vectors = [vector/np.linalg.norm(vector) for vector in vectors]
    outers = [np.outer(a, a) for a in vectors]
    inners = [np.inner(a, a) for a in vectors]
    S = np.zeros((3,3))
    Z = np.sum(inners)
    for s in outers:
        S += s
    S = S/Z
    eig = np.linalg.eig(S)
    evals = eig[0]
    evals.sort()
    return evals, (evals[0]+evals[1])*3/2

def plot_lines(lines):
    xinits = [line[0] for line in lines]
    xfinals = [line[1] for line in lines]
    return plot_segment(xinits, xfinals)

def plot_segment(xinits, xfinals):
    """Plot a list of uniformly generated segments."""
    # xfinals = np.stack(xfinals, axis=-1)
    # # 2-by-3-by-num_segments paths arrays
    # paths = np.stack(xinits, xfinals, axis=0)
    # easier than above, just plot one by one
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    for i in range(len(xinits)):
        xinit = xinits[i]
        xfinal = xfinals[i]
        ax.plot([xinit[0], xfinal[0]], [xinit[1], xfinal[1]], [xinit[2], xfinal[2]])
    ax.set_xlim(0,1)
    ax.set_ylim(0,1)
    ax.set_zlim(0,1)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

def periodize_cylinders(xi, xf, d):
    """A generator for each of the copies of the cylinder that we have to check
    for collisions."""
    # of course the cylinder itself must be used, so we return that first in
    # case anybody only cares to do one, then they do the most important one
    yield xi, xf
    # now for each edge adjacent to the face that the tip of the line defining
    # the center of the cylinder is hitting, we can check if the cylinder
    # spills over onto the next face in that direction by just checking if the
    # points of closest approach from the cylinder to the edge are within r of
    # each other
    xip, xfp = get_true_xif_periodic(xi, xf)
    periodizers = periodizers_given_face[get_face_given_point(xip)]
    for periodizer in periodizers:
        yield periodizer+xip, periodizer+xfp
# old way to do periodization. unfinished. was going to "faster" by manually
# checking to make sure that we had to actually periodize by a face before we
# do. but logic for avoiding repeats,etc. seems hard and it costs a d_line_line
# to do the check anyway, which is the price we pay for just periodizing
# regardless and just making sure it's fine
    # periodization_faces = []
    # for edge_idx in adjacent_edges[xi_face]:
    #     edge = edges[edge_idx]
    #     if d_line_line(*edge, xi, xf):
    #         periodization_faces.append(face_edge_to_face[(xi_face, edge_idx)])

# @jit(nopython=True)
def periodize_point(p):
    periodizers = [-1.0, 0.0, 1.0]
    # points = []
    for offset_x in periodizers:
        for offset_y in periodizers:
            for offset_z in periodizers:
                # points.append(p + np.array([offset_x, offset_y, offset_z]))
                yield p + np.array([offset_x, offset_y, offset_z])
    # return points

# jitting this causes a memory leak
# @jit(nopython=True)
def detect_collision_spheres(c, r, c_new, r_new, periodic):
    if not periodic:
        raise NotImplementedError('collision detection for non-periodic spheres')
    for cp in periodize_point(c):
        for c_newp in periodize_point(c_new):
            if np.linalg.norm(cp - c_newp) < r + r_new:
                return True
    return False

def detect_collision_cylinders(xi, xf, xi_new, xf_new, d, periodic):
    """Are cylinders defined by start and end centerpoint sof xi,xf and
    xi_new,xf_new, each of diameter d, overlapping? use periodic boundary
    conditions in the box of side length a centered in the first quadrant if
    requested."""
    if not periodic:
        return d_segment_segment(xi, xf, xi_new, xf_new) < d
    # one for each of the 26 cubes adjacent to you that can affect you
    else:
        for xi_p, xf_p in periodize_cylinders(xi, xf, d):
            for xi_new_p, xf_new_p in periodize_cylinders(xi_new, xf_new, d):
                # not accurate if POCA is at tip of on or both cylinders
                if d_segment_segment(xi_p, xf_p, xi_new_p, xf_new_p) < d:
                    return True
        return False


# alternate collision detection, more accurate:
# first, if the point of closest approach of the cylinders' centers is interior
# to the cube, then the #TODO

def monte_carlo_with_removal(nsteps, mu, shape_params, shape, objects=None, phi=None,
                             periodic=True):
    """Perform nsteps MC steps. Half steps try to remove a uniformly random
    line with probability exp(-mu). Other half attempts (nsteps times)
    to insert a random line into a box of side length
    a. If the line overlaps with any other line (in the sense that they
    represent the centers of cylinders of radius d), reject the new line.

    Returns table of line insertion attempts and stats about the box at that
    time with a column specifying whether or not the move was accepted.

    TODO: calculate actual volume being added to box instead of just using the
    length*d^2*pi approximation, which only works for thin d."""
    if objects is None:
        objects = []
    num_objects = len(objects)
    if phi is None:
        phi = 0 # volume fraction of cylinders in box
    columns = columns_for_shape(shape)
    move_history = pd.DataFrame(index=np.arange(0, nsteps, 1), columns=columns)
    for i in range(nsteps):
        # two monte carlo moves available: insertion and removal
        if np.random.rand(1) < 0.5: # removal MC move
            # if no object exist, doens't make sense to try to remove them
            if num_objects == 0:
                move_history.loc[i] = row_for_shape(obj=None, shape=shape,
                        phi=phi, num_objects=num_objects, did_succeed=False,
                        is_removal=True)
                continue
            # choose a random object in the box
            remove_ind = int(np.random.rand(1)*num_objects)
            obj_rem = objects[remove_ind]
            # remove it with probability given by chemical potential "mu"
            # according to some parameter of the shape
            L = get_removal_amount(obj_rem, shape)
            did_succeed = np.random.rand(1) < np.exp(-mu*L)
            move_history.loc[i] = row_for_shape(obj_rem, shape, phi, num_objects,
                    did_succeed, is_removal=True)
            if did_succeed:
                objects.pop(remove_ind)
                num_objects -= 1
                phi -= volume_of_shape(obj_rem, shape_params, shape, periodic)
        else: # insertion MC move
            new_obj = new_object(shape_params, shape)
            did_succeed = True
            for existing_obj in objects:
                if detect_collision(existing_obj, new_obj, shape_params, shape, periodic):
                    did_succeed = False
                    break
            move_history.loc[i] = row_for_shape(new_obj, shape, phi, num_objects,
                    did_succeed, is_removal=False)
            if did_succeed:
                num_objects += 1
                phi += volume_of_shape(new_obj, shape_params, shape, periodic)
                objects += [new_obj]
    # the simulation is now done
    # hack in the correct types to the pandas dataframe holding the history fo
    # the moves we've done, since this can't be done at instantiation
    # (open bug in pandas as of 2017-02-24)
    numeric_columns = numeric_columns_for_shape(shape)
    move_history[numeric_columns] = move_history[numeric_columns].apply(pd.to_numeric)
    move_history['did_succeed'] = move_history.did_succeed.astype(bool)
    move_history['is_removal'] = move_history.is_removal.astype(bool)
    # return account of simulation as well as everything you need to continue
    # the simulation where it left off
    return objects, phi, move_history

#DEPRECATED!
# def monte_carlo(nsteps, d, a=1, lines=None, phi=None):
#     """Attempt (nsteps times) to put a random lines into a box of side length
#     a. If the line overlaps with any other line (in the sense that they
#     represent the centers of cylinders of radius d), reject the new line.

#     Returns table of line insertion attempts and stats about the box at that
#     time with a column specifying whether or not the move was accepted.

#     TODO: calculate actual volume being added to box instead of just using the
#     length*d^2*pi approximation, which only works for thin d."""
#     if lines is None:
#         lines = []
#     num_objects = len(lines)
#     if phi is None:
#         phi = 0 # volume fraction of cylinders in box
#     move_history = pd.DataFrame(index=np.arange(0, nsteps, 1),
#                                 columns=['xi1', 'xi2', 'xi3', 'xf1', 'xf2',
#                                          'xf3', 'phi', 'num_objects',
#                                          'did_succeed'])
#     for i in range(nsteps):
#         xi_new, xf_new = uniform_segment_from_cube(a)
#         did_succeed = True
#         for xi, xf in lines:
#             if d_segment_segment(xi, xf, xi_new, xf_new) < d:
#                 did_succeed = False
#                 break
#         move_history.loc[i] = [xi_new[0], xi_new[1], xi_new[2],
#                                xf_new[0], xf_new[1], xf_new[2],
#                                phi, num_objects, did_succeed]
#         if did_succeed:
#             num_objects += 1
#             phi += np.pi*d*d*np.linalg.norm(xf_new - xi_new)
#             lines += [(xi_new, xf_new)]
#     move_history = move_history.apply(pd.to_numeric)

#     return lines, phi, move_history

# this kind of refactoring is not nice because we can't save all the phi's in
# move_history this way
# def try_add_uniform_segment(lines, a=1):
#     xi_new, xf_new = uniform_segment_from_cube(a)
#     did_succeed = True
#     for xi, xf in lines:
#         if d_segment_segment(xi, xf, xi_new, xf_new) < d:
#             did_succeed = False
#             break
#     move_history.loc[i] = [xi_new[0], xi_new[1], xi_new[2],
#                             xf_new[0], xf_new[1], xf_new[2],
#                             phi, num_objects, did_succeed]
#     if did_succeed:
#         line = (xi_new, xf_new)
#         lines += [line]
#         return line
#     return None
def get_all_line_lengths(moves_df):
    return np.sqrt(
        np.power(moves_df['xf3'] - moves_df['xi3'], 2)
      + np.power(moves_df['xf2'] - moves_df['xi2'], 2)
      + np.power(moves_df['xf1'] - moves_df['xi1'], 2)
    )

# original function to run a bunch of sims to get the dependence of the
# acceptance probabilities on phi. we will instead be using the equilibrium
# values from an insertion/removal process to get the acceptance probabilities
# at equilibrium. if you want to come back and still do this, you should modify
# this function to work with the new monte_carlo functions
def get_accept_prob(outdir, nPhi, nL, steps_per_batch, mu,
                    shape_params, shape, num_sims=None):
    """Bin acceptance ratio dependence on phi into nP bins. Restart the
    MC sim every time that the rate that phi is increasing per nS steps is less
    than restart_thresh."""
    if steps_per_batch < 100:
        Warning('nS < 100 will result in bad averaging.')
    if nPhi < 2:
        raise ValueError('nPhi < 2 corresponds to no binning!')
    if nL < 2:
        raise ValueError('nL < 2 corresponds to no binning!')
    # get directory to save to
    outdir = wpath.get_unique_folder(outdir)
    # save index, currently not used
    num_restarts = 0
    save_ind = 0
    # file names to save to
    prob_file = "prob.txt"
    prob_file = os.path.join(outdir, prob_file)
    acc_file = "acc.txt"
    acc_file = os.path.join(outdir, acc_file)
    tot_file = "tot.txt"
    tot_file = os.path.join(outdir, tot_file)
    # initialize bin stuffs
    phiBins = np.linspace(0, 1, nPhi+1)
    phibins_file = os.path.join(outdir, "phibins.txt")
    np.savetxt(phibins_file, phiBins)
    LBins = np.linspace(0, 1*np.sqrt(3), nL+1)
    lbins_file = os.path.join(outdir, "lbins.txt")
    np.savetxt(lbins_file, LBins)
    accepted_moves = np.zeros((nPhi, nL))
    total_moves = np.zeros((nPhi, nL))
    # initialize inter-mc save state
    lines = None
    phi = None
    # initialize restart checker so that we always do a second simulation
    prev_phi_ave = -float('inf')
    while True:
        # perform MC
        lines, phi, new_moves = monte_carlo_with_removal(steps_per_batch,
                mu, shape_params, shape, lines, phi)
        new_moves['L'] = get_all_line_lengths(new_moves)
        # update ratios
        #TODO: check if np.digitize is faster
        for i in range(nPhi):
            for j in range(nL):
                tot_mask = ((phiBins[i] < new_moves['phi'])
                        & (new_moves['phi'] < phiBins[i+1])
                        & (LBins[j] < new_moves['L'])
                        & (new_moves['L'] < LBins[j+1])
                        & (~new_moves['is_removal']))
                acc_mask = tot_mask & (new_moves['did_succeed'])
                accepted_moves[i,j] += np.sum(acc_mask)
                total_moves[i,j] += np.sum(tot_mask)
        # check if should restart
        phi_ave = new_moves['phi'].mean()
        if phi_ave < prev_phi_ave:
            # reset mc state
            lines = None
            phi = None
            # reset to base case
            prev_phi_ave = -float('inf')
            num_restarts += 1
        else:
            # keep track fo previous phi for convergence criterion
            prev_phi_ave = phi_ave
        # write out current values for rate
        np.savetxt(prob_file, np.divide(accepted_moves, total_moves))
        np.savetxt(acc_file, accepted_moves)
        np.savetxt(tot_file, total_moves)
        save_ind += 1
        if num_sims is not None and num_restarts >= num_sims:
            break
    return accepted_moves, total_moves

def get_equilibrium(steps_per_batch, mu, shape_params, shape, lines=None,
                    phi=None, periodic=True):
    """For the values of d and mu provide, run until equilibrium. Here,
    equilibrium means run in batches of steps_per_batch MC steps, then if the
    average phi (packing fraction) in a batch drops below the average phi of
    the previous batch. One more batch is then run to calculate the equilibrium
    value.

    This works since the approach to equlibrium is an exponential (monotone)
    plus noise, so as long as steps_per_batch is chosen large enough that the
    reandom fluctuations during the rise to equilibrium get averaged out.

    For cylinder packign of radius 0.01 in a box of side length 1, a reasonable
    number for steps_per_batch is 1000. For all values tested that return
    reasonable answers, a reasonable number for steps_per_batch is 10000.
    """
    prev_phi_ave = -float('inf')
    lines = []; phi = 0;
    while True:
        lines, phi, moves = monte_carlo_with_removal(steps_per_batch, mu,
                shape_params, shape, lines, phi, periodic=periodic)
        moves = moves[~moves.is_removal]
        phi_ave = moves['phi'].mean()
        # if termination condition reached, run again so taht we decorrelate
        # from the artificial lower value of phi
        if phi_ave < prev_phi_ave:
            _, _, moves = monte_carlo_with_removal(steps_per_batch, mu,
                    shape_params, shape, lines, phi, periodic=periodic)
            moves = moves[~moves.is_removal]
            return moves['phi'].mean()
        prev_phi_ave = phi_ave


def equilibrium_mapper(p):
    return (p['s'], p['mu'], p['shape_param'],
            get_equilibrium(p['s'], p['mu'], p['shape_param'], p['shape'], periodic=p['p']))

def scan_equilibriums(steps_per_batch, mus, shape_params, shape, periodic=True, num_cores=8):
    """Use pscan to get a list of equilibrium phi levels as a function of the
    removal probability (i.e. mu a.k.a. chemical potential of the bath of
    "sticks" feeding our box filling process) and the diameter/box width ratio.

    steps_per_batch must be scalar, or same size as mus, they'll be
    jparams.
    """
    script_name = os.path.basename(__file__)
    #print(script_name + ': Running scan_equilibriums!')
    p = multiprocessing.Pool(num_cores)
    if type(steps_per_batch) is int:
        steps_per_batch = [steps_per_batch]
    if type(shape) is str:
        shape = [shape]
    if len(steps_per_batch) == 1:
        steps_per_batch = steps_per_batch*np.ones_like(mus, dtype=np.dtype(int))
    jparam = {'s': steps_per_batch, 'mu': mus}
    scan = pscan.Scan({'p': [periodic], 'shape_param': shape_params,
            'shape': shape})
    scan.add_jparam(jparam)
    print("mu\twidth_ratio\tequilibrium_phi\tnsteps")
    # for s,mu,sp,phi in map(equilibrium_mapper, scan.params()):
    for s,mu,sp,phi in p.imap_unordered(
            equilibrium_mapper, scan.params(), chunksize=10):
        print(str(mu) + '\t' + str(sp) + '\t' + str(phi) + '\t' + str(s))

def plot_equilibrium_scan(outfile):
    df = pd.read_table(outfile)
    hf = plt.figure()
    ha = hf.add_subplot(111, projection='3d')
    ha.scatter(df.width_ratio, df.equilibrium_phi, df.mu)
    ha.set_xlabel('$d$: Cylinder diameter/Box side length')
    ha.set_ylabel('$\phi$: Equilibrium packing density')
    ha.set_zlabel('$\mu$: Chemical potential/length')
    ha.set_title('Equilibrium Packing Densities from MC')
    return ha

# approx packing fract of cylinders is 0.73
def balls_approximation(phi, N, phi_full=0.73):
    phi_solvent = 1 - phi/phi_full
    return N*( phi_solvent*np.log(phi_solvent) - phi_solvent + 1 )

def fit_equilibrium_scan(outfile):
    """Try to get a fit for mu as a function of d and phi. So far, we tried
    "balls_approximation" above, which Quinn derived just based on
    inserting a comparable number of identifiable beads into a bin. but that
    function is concave up in planes of fixed d, and ours is concave down
    for the first long time before it slopes up rapidly (presumedly due to
    nematic effects, should check this)."""
    df = pd.read_table(outfile)
    ds = df.width_ratio.unique()
    ds.sort()
    for d in ds:
        dfd = df[df.width_ratio == d]
        phi = np.array(dfd.equilibrium_phi)
        mu = np.array(dfd.mu)
        popt, pcov = curve_fit(balls_approximation, phi, mu, p0=[d*d*d])
        import pdb; pdb.set_trace()



def plot_accept_prob(outdir):
    files = os.listdir(outdir)
    prob = None; acc = None; tot = None
    phiBins = None; Lbins = None
    for file in files:
        if file.startswith('prob'):
            prob = np.loadtxt(os.path.join(outdir, file))
        elif file.startswith('acc'):
            acc = np.loadtxt(os.path.join(outdir, file))
        elif file.startswith('tot'):
            tot = np.loadtxt(os.path.join(outdir, file))
        elif file.startswith('phibins'):
            phiBins = np.loadtxt(os.path.join(outdir, file))
        elif file.startswith('lbins'):
            Lbins = np.loadtxt(os.path.join(outdir, file))

    if prob is None or acc is None or tot is None \
            or phiBins is None or Lbins is None:
        raise FileNotFoundError('one of [phi/l]bins, prob, acc, tot not found in ' + outdir)
    # due to a bug, accepted rates not seem to be saved correctly
    acc = tot*prob
    # get bin centers for plotting
    dphiBin = phiBins[1] - phiBins[0]
    phiBinC = phiBins[:-1] + dphiBin/2
    dLbin = Lbins[1] - Lbins[0]
    LbinC = Lbins[:-1] + dphiBin/2
    phimesh = np.zeros((len(phiBinC), len(LbinC)))
    Lmesh = np.zeros((len(phiBinC), len(LbinC)))
    for i in range(len(phiBinC)):
        for j in range(len(LbinC)):
            phimesh[i,j] = phiBinC[i]
            Lmesh[i,j] = LbinC[j]
    X = phimesh; Y = Lmesh; Z = -np.log(prob)/Lmesh;
    Zave = -np.log(prob) # *Lmesh/Lmesh
    is_missing = np.isnan(Zave) | np.isinf(Zave)
    Zave[is_missing] = 0.0
    Lnorm = Lmesh;
    Lnorm[is_missing] = 0.0
    Zave = np.sum(Zave, axis=1)/np.sum(Lnorm, axis=1);
    hf = plt.figure()
    ha = hf.add_subplot(111, projection='3d')
    ha.scatter(X,Y,Z)
    ha.set_xlabel('$\phi$: Density Occupied')
    ha.set_ylabel('$\Delta{}L$: length of inserted stick')
    ha.set_zlabel('-ln[Prob(Accept)]/$\Delta{}L$')
    hf2 = plt.figure()
    ha2 = hf2.add_subplot(111)
    plt.scatter(X,Z,s=1,c=Y)
    plt.colorbar(label='$\Delta{}L')
    ha2.set_xlabel('$\phi$: Density Occupied')
    ha2.set_ylabel('-ln[Prob(Accept)]/$\Delta{}L$')
    plt.plot(X,Zave)
    return ha,ha2

def get_mu_of_d_phi(equilibrium_outfile, nphigrid=100):
    df = pd.read_table(equilibrium_outfile)
    dgrid = df.width_ratio.unique()
    dgrid.sort()
    phigrid = np.linspace(0, 1, nphigrid)
    mu = np.zeros((len(dgrid), nphigrid))
    mu[:,:] = np.nan
    zs = {}
    for i,d in enumerate(dgrid):
        lowess = sm.nonparametric.lowess
        dfd = df[df.width_ratio == d]
        x = dfd.equilibrium_phi
        y = dfd.mu
        z = lowess(y, x, frac=1.0/3.0)
        zs[d] = z
        xfit = z[:,0]
        yfit = z[:,1]
        # this should in principle be faster and generalize to arbitrary
        # dimensions, but DEFINITELY doesn't work as is. not sure what's
        # wrong. gives "almost" correct results
        # tree = cKDTree(np.expand_dims(xfit, axis=1))
        # dist, ind = tree.query(np.expand_dims(phigrid, axis=1), k=2)
        # d1, d2 = dist.T
        # x1, x2 = xfit[ind].T
        # v1, v2 = yfit[ind].T
        # mult = np.ones_like(v1)
        # mult[x1 > x2] = -1.0
        # mu[i,:] = mult*(d1)/(d1 + d2)*(v2 - v1) + v1
        # mu[i, phigrid < xfit[0]] = -float('inf')
        # mu[i, phigrid > xfit[-1]] = float('inf')
        for j,phi in enumerate(phigrid):
            inds = np.argwhere(xfit < phi)
            if inds.size == 0:
                mu[i,j] = -float('inf')
                continue
            ind_l = inds[-1]
            inds = np.argwhere(xfit > phi)
            if inds.size == 0:
                mu[i,j] = float('inf')
                continue
            ind_r = inds[0]
            x1 = xfit[ind_l]
            x2 = xfit[ind_r]
            y1 = yfit[ind_l]
            y2 = yfit[ind_r]
            mu[i,j] = y1 + (phi - x1)*(y2 - y1)/(x2 - x1)
    return dgrid, phigrid, mu

def plot_equilibrium_scan_with_interp(outfile, *args, **kwargs):
    ha = plot_equilibrium_scan(outfile)
    dgrid, phigrid, mu = get_mu_of_d_phi(outfile, *args, **kwargs)
    X, Y = np.meshgrid(dgrid, phigrid)
    ha.plot_surface(X, Y, mu.T, alpha=0.5)
    return ha

if __name__ == '__main__':

    # nsteps = np.linspace(1000, 5000, 101).astype(int)
    nsteps = 2000
    mus = np.flipud(np.linspace(0.01, 10, 101))
    ds = np.flipud(np.linspace(0.01, 0.3, 31))
    scan_equilibriums(nsteps, mus, ds, 'sphere', num_cores=32, periodic=True)

    # get_accept_prob('tmp/accept_probs', nPhi=100, nL=50,
    #                 steps_per_batch=1000, mu=5, d=0.01,
    #                 num_sims=1000)

    # lines, phi, move_history = monte_carlo_with_removal(10000, 1, 0.1, periodic=True)
    # print(move_history.phi.mean())






"""
wlcsim.tests contains functions for validating the output of wlcsim.

For now, most of these are just wrappers to the functions used in Quinn and
Shifan's simulations (SpakowitzLab/MCparallelMPI/utility/*)
"""

from pymatbridge import Matlab
import os
import .data as wdata

def test_init_end_to_end_distance(run_dir):
    """
    Plot the end-to-end distance distribution for a set of runs' r0 versus the
    theoretical distribution.
    """
    if not os.path.isdir(os.path.join(run_dir, 'data', 'end_to_end_r0')):
        wdata.aggregate_r0_group_NP_L0(run_dir)
    m = Matlab()
    m.start()
    m.run_code("help pcalc")
    m.stop()




import numpy as np
import pdb
from scipy.linalg import expm
import matplotlib
import matplotlib.pyplot as plt

from .yyyreal import YYY
def alpha(l,m):
    return np.sqrt((3*(l-m)*(l+m))/(4*np.pi*(2*l-1)*(2*l+1)))
Y00 = 1.0/np.sqrt(4*np.pi)

lmax = 50


def getM2(m):
    M = np.zeros((lmax+1, lmax+1))

    for l in range(abs(m)+2, lmax+1):
        M[l, l-2] = alpha(l, m)*alpha(l-1, m)/alpha(2, 0)
    for l in range(abs(m), lmax+1):
        M[l, l] = (alpha(l+1, m)**2 - alpha(1, 0)*Y00 +alpha(l, m)**2)/alpha(2, 0)
    for l in range(abs(m), lmax+1-2):
        M[l, l+2] = alpha(l+1, m)*alpha(l+2, m)/alpha(2, 0)

    return M

M2= {}
for m in [-2, -1, 0, 1, 2]:
    M2[m] = getM2(m)


def rodFun(m_1, m_2, m_k_1, m_k_2, eM):
    const = (2*np.pi/3.0)*(eM[0][0,2]**2)/(eM[m_1][2,2]*eM[m_2][2,2]*eM[0][0,0])
    summ = 0.0

    # Calculate M's allowable by selection rules
    if m_k_1==0:
        Mset = {m_1}
    elif m_k_1==-1:
        Mset = {-m_1+1, -m_1-1}
    elif m_k_1==1:
        Mset = {m_1+1, m_1-1}
    if m_k_2==0:
        Mset = Mset.intersection({m_2})
    elif m_k_2==-1:
        Mset = Mset.intersection({-m_2-1,-m_2+1})
    elif m_k_2==1:
        Mset = Mset.intersection({m_2-1, m_2+1})

    for M in Mset:
        summ = summ + (YYY(m_1, 1, m_k_1, M, ORDER_L=lmax+1)@eM[M]
                       @YYY(M, 1, m_k_2, m_2, ORDER_L=lmax+1))[2,2]
    return const*summ





basis = [(-1,-1), (-1,1), (0,-1), (0,1), (1,-1), (1,1)] # (m_k_1, m_1)

def showMatrix(gamma=1.2345):
    out = np.zeros((6,6))
    np.set_printoptions(precision=5)
    np.set_printoptions(linewidth=200)

    eM = {}
    for m in [-2,-1,0,1,2]:
        eM[m] = expm( gamma*M2[m])
    print("eM[-2]")
    print(eM[-2][0:6,0:6])
    print("eM[-1]")
    print(eM[-1][0:6,0:6])
    print("eM[1]")
    print(eM[1][0:6,0:6])
    print("eM[2]")
    print(eM[2][0:6,0:6])

    for ii, (m_k_1, m_1) in enumerate(basis):
        for jj, (m_k_2, m_2) in enumerate(basis):
            out[ii,jj] = rodFun(m_1, m_2, m_k_1, m_k_2, eM)

    print("Frank Matrix")
    print(out)
#showMatrix()

def RR_datapoint(gamma):
    """Takes what is considered gamma*N for WLC
    """
    eM = {}
    for m in [-2,-1,0,1,2]:
        eM[m] = expm( gamma*M2[m])

    K_bend = rodFun(1, 1, 0, 0, eM)
    K_twist = rodFun(-1, -1, 1, 1, eM)
    K_splay = rodFun(1, 1, 1, 1, eM)
    a_prime = 15*gamma*eM[0][0,0]/(4*np.sqrt(np.pi)*eM[0][0,2])
    return {'aLAf':a_prime, "K'bend":K_bend, "K'twist":K_twist, "K'splay":K_splay}

def rodFrank(gammas=None, plot=False): 
    if gammas is None:
        npts=200
        gammas = np.linspace(0.0,50.0,npts)
    else:
        npts=len(gammas)
    K_bend = np.zeros(npts)
    K_twist = np.zeros(npts)
    K_splay = np.zeros(npts)
    a_prime = np.zeros(npts)
    
    for ii in range(npts):
        eM = {}
        for m in [-2,-1,0,1,2]:
            eM[m] = expm( gammas[ii]*M2[m])

        K_bend[ii] = rodFun(1, 1, 0, 0, eM)
        K_twist[ii] = rodFun(-1, -1, 1, 1, eM)
        K_splay[ii] = rodFun(1, 1, 1, 1, eM)
        
        a_prime[ii] = 15*gammas[ii]*eM[0][0,0]/(4*np.sqrt(np.pi)*eM[0][0,2])
   
    out = {"K_bend":K_bend, "K_twist":K_twist, "K_splay":K_splay}
    temp=[]
    for idx, val in enumerate(a_prime):
        if np.isnan(val):
            continue
        temp.append((val,idx))
    val, idx = min(temp)

    out["a_min"] = val
    out["idx_a_min"] = idx
    out["gammas*N"] = gammas
    out["aLAf"] = a_prime


    if plot:
        matplotlib.rcParams['text.usetex'] = True
        font = {'size':16}
        matplotlib.rc('font', **font)
        fig, ax = plt.subplots(tight_layout=True)
        ax.plot(gammas, K_bend, label="$K_{bend}$")
        ax.plot(gammas, K_twist, label="$K_{twist}$")
        ax.plot(gammas, K_splay, label="$K_{splay}$")
        plt.xlabel(r'$ \gamma $')
        plt.ylabel(r'$ K\left(\frac{A}{L}\frac{V}{nAL}\right)$ ')
        #plt.ylabel("K'")
        plt.legend()
        plt.show()

        plt.plot(gammas, a_prime)
        plt.title(r"$a\left(\frac{nA^2 L^2}{V}\right)$")
        plt.xlabel(r'$\gamma$')
        plt.tight_layout()
        plt.show()

        fig, ax = plt.subplots(tight_layout=True)
        ax.plot(a_prime, K_bend, label="$K_{bend}$")
        ax.plot(a_prime, K_twist, label="$K_{twist}$")
        ax.plot(a_prime, K_splay, label="$K_{splay}$")
        plt.xlabel(r"$a\left(\frac{nA^2 L^2}{V}\right)$")
        plt.ylabel(r'$ K\left(\frac{A}{L}\frac{V}{nAL}\right)$ ')
        #plt.ylabel("K'")
        plt.legend()
        plt.tight_layout()
        plt.show()
    else:
        return out

#rodFrank(plot=True)
    




import numpy as np

from functools import reduce


def center_by_mass(x, particle_axis=0):
    """Subtract center of mass (unweighted) from a collection of vectors
    corresponding to particles's coordinates."""
    shape = x.shape
    # center of mass is the average position of the particles, so average over
    # the particles
    centers_of_mass = np.mean(x, particle_axis)
    # tile the centers of mass for clean elementwise division
    # is there a way to do this faster by relying on numpy's array projection
    # semantics?
    tile_shape = [shape[i] if i == particle_axis else 1
                  for i in range(len(shape))]
    centers_of_mass = np.tile(centers_of_mass, tile_shape)
    return x - centers_of_mass

from scipy import arange, pi, sqrt, zeros
from scipy.special import spherical_jn
from scipy.optimize import brentq
def Jn(r,n):
    return spherical_jn(n, r)

def spherical_jn_zeros(n, nt):
    """recursive method: computes zeros ranges of sphereical_jn(r,n) from zeros
    of sphereical_jj(r,n-1)

    pros : you are certain to find the right zeros values;
    cons : all zeros of the n-1 previous Jn have to be computed;
    note : Jn(r,0) = sin(r)/r
    """
    zerosj = zeros((n+1, nt), dtype=float)
    zerosj[0] = arange(1,nt+1)*pi
    points = arange(1,nt+n+1)*pi
    racines = zeros(nt+n, dtype=float)
    for i in range(1,n+1):
        for j in range(nt+n-i):
            foo = brentq(Jn, points[j], points[j+1], (i,))
            racines[j] = foo
        points = racines
        zerosj[i][:nt] = racines[:nt]
    return (zerosj)



"""Rouse polymer, analytical results.

Notes
-----
There are two parameterizations of the "Rouse" polymer that are commonly used,
and they use the same variable name for two different things.

In one, N is the number of Kuhn lengths, and in the other, N is the number of
beads, each of which can represent an arbitrary number of Kuhn lengths.
"""
from ..utils.math import spherical_jn_zeros
from bruno_util.mittag_leffler import ml as mittag_leffler

import numpy as np
import scipy
import scipy.special
import spycial
from scipy.signal import savgol_filter, savgol_coeffs
from numba import jit
import mpmath

from functools import lru_cache
from pathlib import Path
import os

_default_modes = 10000

@jit
def rouse_mode(p, n, N=1):
    """Eigenbasis for Rouse model.

    Indexed by p, depends only on position n/N along the polymer of length N.
    N=1 by default.

    Weber, Phys Rev E, 2010 (Eq 14)"""
    p = np.atleast_1d(p)
    phi = np.sqrt(2)*np.cos(p*np.pi*n/N)
    phi[p == 0] = 1
    return phi

@jit(nopython=True)
def rouse_mode_coef(p, b, N, kbT=1):
    """k_p: Weber Phys Rev E 2010, after Eq. 18."""
    # alternate: k*pi**2/N * p**2, i.e. k = 3kbT/b**2
    return 3*np.pi**2*kbT/(N*b**2)*p**2

@jit(nopython=True)
def kp_over_kbt(p : float, b : float, N : float):
    """k_p/(k_B T) : "non-dimensionalized" k_p is all that's needed for most
    formulas, e.g. MSD."""
    return (3*np.pi*np.pi)/(N*b*b) * p*p

@jit(nopython=True)
def linear_mid_msd(t, b, N, D, num_modes=_default_modes):
    """
    modified from Weber Phys Rev E 2010, Eq. 24.
    """
    rouse_corr = np.zeros_like(t)
    for p in range(1, num_modes+1):
        # k2p = rouse_mode_coef(2*p, b, N, kbT)
        # rouse_corr += 12*kbT/k2p*(1 - np.exp(-k2p*t/(N*xi)))
        k2p_norm = kp_over_kbt(2*p, b, N)
        rouse_corr += (1/k2p_norm)*(1 - np.exp(-k2p_norm*(D/N)*t))
    # return rouse_corr + 6*kbT/xi/N*t
    return 12*rouse_corr + 6*D*t/N

def rouse_large_cvv_g(t, delta, deltaN, b, D):
    """Cvv^delta(t) for infinite polymer.

    Lampo, BPJ, 2016 Eq. 16."""
    # k = 3*kbT/b**2
    ndmap = lambda G, arr: np.array(list(map(G, arr)))
    G = lambda x: float(mpmath.meijerg([[],[3/2]], [[0,1/2],[]], x))
    # we can use the fact that :math:`k/\xi = 3D/b^2` to replace
    # gtmd = ndmap(G, np.power(deltaN, 2)*xi/(4*k*np.abs(t-delta)))
    # gtpd = ndmap(G, np.power(deltaN, 2)*xi/(4*k*np.abs(t+delta)))
    # gt = ndmap(G, np.power(deltaN, 2)*xi/(4*k*np.abs(t)))
    # with the same formulas in terms of "D"
    gtmd = ndmap(G, np.power(deltaN, 2)*b*b/(4*3*D*np.abs(t-delta)))
    gtpd = ndmap(G, np.power(deltaN, 2)*b*b/(4*3*D*np.abs(t+delta)))
    gt = ndmap(G, np.power(deltaN, 2)*b*b/(4*3*D*np.abs(t)))
    # tricky conversion from Tom's formula
    # :math:`k_BT / \sqrt{\xi*k} = \frac{k_BT}{\xi} / \sqrt{k / \xi}`
    # and we know :math:`k/\xi = \frac{3D}{b^2}`.
    # so when the dust settles,
    # :math:`\frac{3k_BT}{\delta^2\sqrt{\xi k}} = \frac{b\sqrt{3D}}{\delta^2}
    # so instead of the expression from Tom's paper:
    # return 3*kbT/(np.power(delta, 2)*np.sqrt(xi*k)) * (
    #     np.power(np.abs(t - delta), 1/2)*gtmd
    #   + np.power(np.abs(t + delta), 1/2)*gtpd
    #   - 2*np.power(np.abs(t), 1/2)*gt
    # )
    # we can simply return
    return b*np.sqrt(3*D)*np.power(delta, -2) * (
        np.power(np.abs(t - delta), 1/2)*gtmd
      + np.power(np.abs(t + delta), 1/2)*gtpd
      - 2*np.power(np.abs(t), 1/2)*gt
    )


mod_file = os.path.abspath(__file__)
mod_path = os.path.dirname(mod_file)


def end2end_distance(r, lp, N, L):
    """
    For now, always returns values for ``r = np.linspace(0, 1, 50001)``.

    Parameters
    ----------
    r : (N, ) float, array_like
        the values at which to evaluate the end-to-end probability
        distribution. ignored for now (TODO: fix)
    lp : float
        persistence length of polymer
    N : int
        number of beads in polymer
    L : float
        polymer length

    Returns
    -------
    x : (5001,) float
        ``np.linspace(0, 1, 5001)``
    g : (5001,) float
        :math:`P(|R| = x | lp, N, L)`

    Notes
    -----
    Uses the gaussian chain whenever applicable, ssWLC tabulated values
    otherwise. If you request parameters that require a WLC end-to-end
    distance, the function will ValueError.
    """
    Delta = L/(lp*(N-1))  # as in wlcsim code
    # WLC case
    actual_r = np.linspace(0, 1, 5001)
    if Delta < 0.01:  # as in wlcsim code
        ValueError('end2end_distance: doesn\'t know how to compute WLC case!')
    # ssWLC case
    elif Delta < 10:  # as in wlcsim code
        Eps = N/(2*lp*L)  # as in wlcsim code
        file_num = round(100*Eps)  # index into file-wise tabulated values
        file_name = os.path.join('pdata', 'out' + str(file_num) + '.txt')
        file_name = os.path.join(mod_path, file_name)
        G = np.loadtxt(file_name)
        return (actual_r, G)
    # GC case
    else:
        return (actual_r, end2end_distance_gauss(actual_r, b=2*lp, N=N, L=L))

def end2end_distance_gauss(r, b, N, L):
    """ in each dimension... ? seems to be off by a factor of 3 from the
    simulation...."""
    r2 = np.power(r, 2)
    return 3.0*r2*np.sqrt(6/np.pi)*np.power(N/(b*L), 1.5) \
            *np.exp(-(3/2)*(N/(b*L))*r2)

def test_rouse_msd_line_approx():
    """Figure out what Dapp is exactly using our analytical result.

    if we use msd_approx(t) = 3*bhat*np.sqrt(Dhat*t)/np.sqrt(3)*1.1283791(6)
    then ``|msd(t) - msd_approx(t)|/msd(t) = np.sqrt(2)/N*np.power(t, -1/2)``.
    ``msd_approx(t) > msd(t)`` in this range, so that means that
    ``msd_approx(t)/(np.sqrt(2)/N*np.power(t, -1/2) + 1) = msd(t)``.

    if we redefine msd_approx with this correction, the new relative error is
    about ``np.sqrt(2)/N/100*t**(-1/2)``? gotta see if this carries over to
    other chain parameters though...it does not...  this time msd(t) is bigger,
    so ``msd_approx = msd_approx(t)/(1 - np.sqrt(2)/N/100*t**(-1/2))``.

    okay actually looks like extra factor is additive?

    for N = 1e8+1; L = 25; b = 2; D = 1;, we have
    msd_approx(t) - msd(t) = 1.01321183e-07

    for N = 1e8+1; L = 174; b = 150; D = 166;, we have
    msd_approx(t) - msd(t) = 5.2889657(3)e-05

    for N = 1e8+1; L = 174; b = 15; D = 166;, we have
    msd_approx(t) - msd(t) = 5.2889657(3)e-06

    for N = 1e8+1; L = 17.4; b = 15; D = 166;, we have
    msd_approx(t) - msd(t) = 5.2889657(3)e-07

    for N = 1e8+1; L = 17.4; b = 15; D = 16.6;, we have
    msd_approx(t) - msd(t) = 5.2889657(3)e-07 (no change)

    for N = 1e7+1; L = 17.4; b = 15; D = 16.6;, we have
    msd_approx(t) - msd(t) = 5.2889657(3)e-06 (no change)

    so the answer is like
    3*bhat*np.sqrt(Dhat*t)/np.sqrt(3)*1.1283791615 - 0.202642385398*b*L/N
    """
    N = 1e8+1; L = 174; b = 150; D = 166; dt = 1e-2; Nt = 1e6; Nt_save = 1e4;
    t = np.arange(0, Nt*dt, dt); t_save = t[::int(Nt/Nt_save)];
    Nhat = L/b; L0 = L/(N-1); Dhat = D*(N)/Nhat; bhat = np.sqrt(L0*b)
    plt.plot(tmsd, np.abs(msd_rouse - 3*bhat*np.sqrt(Dhat*tmsd)/np.sqrt(3)*1.12837916)/msd_rouse, 'k')
    plt.yscale('log')
    plt.xscale('log')

@jit(nopython=True)
def gaussian_G(r, N, b):
    """Green's function of a Gaussian chain at N Kuhn lengths of separation,
    given a Kuhn length of b"""
    r2 = np.power(r, 2)
    return np.power(3/(2*np.pi*b*b*N), 3/2)*np.exp(-(3/2)*r2/(N*b*b))

@jit(nopython=True)
def gaussian_Ploop(a, N, b):
    """Looping probability for two loci on a Gaussian chain N kuhn lengths
    apart, when the Kuhn length is b, and the capture radius is a"""
    Nb2 = N*b*b
    return spycial.erf(a*np.sqrt(3/2/Nb2)) - a*np.sqrt(6/np.pi/Nb2)/np.exp(3*a*a/2/Nb2)

@jit(nopython=True)
def _cart_to_sph(x, y, z):
    r = np.sqrt(x*x + y*y + z*z)
    if r == 0.0:
        return 0.0, 0.0, 0.0
    phi = np.arctan2(y, x)
    theta = np.arccos(z/r)
    return r, phi, theta

def confined_G(r, rp, N, b, a, n_max=100, l_max=50):
    # first precompute the zeros of the spherical bessel functions, since our
    # routine to do so is pretty slow
    if l_max >= 86: # for l >= 86, |m| >= 85 returns NaN from sph_harm
        raise ValueError("l_max > 85 causes NaN's from scipy.special.sph_harm")
    if confined_G.zl_n is None or n_max > confined_G.zl_n.shape[1] \
            or l_max > confined_G.zl_n.shape[0]:
        confined_G.zl_n = spherical_jn_zeros(l_max, n_max)
    # some aliases
    spherical_jn = scipy.special.spherical_jn
    Ylm = scipy.special.sph_harm
    zl_n = confined_G.zl_n[:l_max+1,:n_max]
    # convert to spherical coordinates
    r = np.array(r)
    if r.ndim == 1:
        x, y, z = r
        xp, yp, zp = rp
    # elif r.ndim == 2:
    #     x = r[:,0]
    #     y = r[:,1]
    #     z = r[:,2]
    #     xp = rp[:,0]
    #     yp = rp[:,1]
    #     zp = rp[:,2]
    else:
        raise ValueError("Don't understand your R vectors")
    r, phi, theta = _cart_to_sph(x, y, z)
    rp, phip, thetap = _cart_to_sph(xp, yp, zp)
    # compute terms based on indexing (ij in meshgrid to match zl_n)
    l, n = np.meshgrid(np.arange(l_max+1), np.arange(n_max), indexing='ij')
    ln_term = 2*np.exp(-b**2/6 * (zl_n/a)**2 * N)
    ln_term = ln_term/(a**3 * spherical_jn(l+1, zl_n)**2)
    ln_term = ln_term*spherical_jn(l, zl_n/a*r)*spherical_jn(l, zl_n/a*rp)
    # l,m terms
    m = np.arange(-l_max, l_max+1)
    l, m = np.meshgrid(np.arange(l_max+1), np.arange(-l_max, l_max+1))
    lm_term = Ylm(m, l, phi, theta)*Ylm(m, l, phip, thetap)
    lm_mask = np.abs(m) <= l
    lm_term[~lm_mask] = 0
    # now broadcast and sum
    G = np.sum(ln_term[None,:,:] * lm_term[:,:,None])
    return G
confined_G.zl_n = None


def linear_mscd(t, D, Ndel, N, b=1, num_modes=_default_modes):
    r"""
    Compute mscd for two points on a linear polymer.

    Parameters
    ----------
    t : (M,) float, array_like
        Times at which to evaluate the MSCD
    D : float
        Diffusion coefficient, (in desired output length units). Equal to
        :math:`k_BT/\xi` for :math:`\xi` in units of "per Kuhn length".
    Ndel : float
        Distance from the last linkage site to the measured site. This ends up
        being (1/2)*separation between the loci (in Kuhn lengths).
    N : float
        The full lengh of the linear polymer (in Kuhn lengths).
    b : float
        The Kuhn length (in desired length units).
    num_modes : int
        how many Rouse modes to include in the sum

    Returns
    -------
    mscd : (M,) np.array<float>
        result
    """
    mscd = np.zeros_like(t)

    k1 = 3 * np.pi ** 2 / (N * (b ** 2))
    sum_coeff = 48 / k1
    exp_coeff = k1 * D / N
    sin_coeff = np.pi * Ndel / N

    for p in range(1, num_modes+1, 2):
        mscd += (1/p**2) * (1 - np.exp(-exp_coeff * (p ** 2) * t)) \
                * np.sin(sin_coeff*p)**2

    return sum_coeff * mscd


def ring_mscd(t, D, Ndel, N, b=1, num_modes=_default_modes):
    r"""
    Compute mscd for two points on a ring.

    Parameters
    ----------
    t : (M,) float, array_like
        Times at which to evaluate the MSCD.
    D : float
        Diffusion coefficient, (in desired output length units). Equal to
        :math:`k_BT/\xi` for :math:`\xi` in units of "per Kuhn length".
    Ndel : float
        (1/2)*separation between the loci on loop (in Kuhn lengths)
    N : float
        full length of the loop (in Kuhn lengths)
    b : float
        The Kuhn length, in desired output length units.
    num_modes : int
        How many Rouse modes to include in the sum.

    Returns
    -------
    mscd : (M,) np.array<float>
        result
    """
    mscd = np.zeros_like(t)

    k1 = 12 * np.pi ** 2 / (N * (b ** 2))
    sum_coeff = 48 / k1
    exp_coeff = k1 * D / N
    sin_coeff = 2 * np.pi * Ndel / N

    for p in range(1, num_modes+1):
        mscd += (1 / p ** 2) * (1 - np.exp(-exp_coeff * p ** 2 * t)) \
                * np.sin(sin_coeff * p) ** 2
    return sum_coeff * mscd



def end_to_end_corr(t, D, N, num_modes=_default_modes):
    """Doi and Edwards, Eq. 4.35"""
    mscd = np.zeros_like(t)
    tau1 = N**2/(3*np.pi*np.pi*D)
    for p in range(1, num_modes+1, 2):
        mscd += 8/p/p/np.pi/np.pi * np.exp(-t*p**2 / tau1)
    return N*mscd



"""Generate equilibrium conformations of Wormlike Chains (WLCs) exactly."""
from ..utils.rotation import *

import numpy as np
from numpy.random import random_sample as urand
import scipy
import scipy.optimize
from scipy.special import erfi
from numba import jit

from functools import partial
import multiprocessing
from multiprocessing import Pool

proc_max = multiprocessing.cpu_count()

def phi_cdf(phi, eps):
    """The cumulative distribution of a given azimuthal angle :math:`phi` given
    a distance between neighboring beads of :math:`\Epsilon = ds/l_p`."""
    f0, fpi = phi_Z_0_1_(eps)
    return (phi_indef_(phi, eps) - f0)/(fpi - f0)

def phi_indef_(phi, eps):
    """proportional to the indefinite integral of sin(phi)exp(-eps*phi**2/2)"""
    return np.real(
            erfi((1 + 1j*eps*phi)/np.sqrt(2*eps))
          + erfi((1 - 1j*eps*phi)/np.sqrt(2*eps))
    )

def phi_Z_0_1_(eps):
    """return F(0), F(pi), so that (phi_indef_(theta)-F(0))/(F(pi)-F(0)) is a
    CDF."""
    return phi_indef_(0, eps), phi_indef_(np.pi, eps)

def phi_indef_2_(phi, eps):
    """alternate representation of the indefinite integral of
    sin(phi)exp(-phi**2/(2*sigma)), different notation?"""
    return np.real(np.sqrt(np.pi*eps/2)/2*np.exp(-eps/2)*(
        - erfi((eps - 1j*phi)/np.sqrt(2*eps))
        - erfi((eps + 1j*phi)/np.sqrt(2*eps))
    ))

def phi_z_2_(eps):
    """normalization factor. phi_indef_(pi)-phi_indef_(0)"""
    return np.real(np.sqrt(np.pi*eps/2)/2*np.exp(-eps/2)*(
        - erfi((eps - 1j*np.pi)/np.sqrt(2*eps))
        + 2*erfi(np.sqrt(eps/2))
        - erfi((eps + 1j*np.pi)/np.sqrt(2*eps))
    ))

def phi_cdf_2(phi, eps):
    """alternate representation of The cumulative distribution of a given
    azimuthal angle :math:`phi` given a distance between neighboring beads of
    :math:`\Epsilon = ds/l_p`."""
    return (phi_indef_(phi, eps) - phi_indef_(0, eps))/phi_z_(eps)

def phi_prob(phi, eps):
    """The probability distribution of a given azimuthal angle :math:`phi`
    given a distance between neighboring beads of :math:`\Epsilon = ds/l_p`."""
    return np.sin(phi)*np.exp(-np.power(phi, 2)/2/eps)/phi_z_(eps)

def dphi_prob(phi, eps):
    """Derivative of phi_prob w.r.t. phi."""
    dphi = np.cos(phi)*np.exp(-np.power(phi, 2)/2/eps) \
         - np.sin(phi)*np.exp(-np.power(phi, 2)/2/eps)*2*phi/2/eps
    return dphi/phi_z_(eps)


def Fzero_(phi, eps, xi, f0, fpi):
    return (phi_indef_(phi, eps) - f0)/(fpi - f0) - xi
def inv_sample_(xi, eps, f0, fpi):
    return scipy.optimize.toms748(Fzero_, 0, np.pi, k=2, args=(eps, xi, f0, fpi))

def dwlc_k(N, L, lp, n_proc=proc_max-1):
    r"""Draw discrete wormlike chain steps, piecewise constant curvature.

    The energy of the wormlike chain is proportional to the curvature of the
    chain squared. We make the assumption that the curvature is constant, and
    that we are approximating the curve via its osculating circle at each
    point.

    For the discrete WLC, we want the next bead to be on the sphere centered at
    the previous bead with radius equal to the length of chain between the
    beads (we call this eps).

    To decide where on that sphere the next bead should be placed, we write the
    curve as r(s) and define the azimuthal angle between the two beads to be
    phi(ds) = arccos(u(s) . u(s+ds)), where ds is the path length between the
    beads. Assuming constant curvature between the two beads, we have simply
    that phi(ds) = kappa*ds. Thus, the probability of a given azimuthal angle
    is just

    .. math::

        P(\phi | \theta) \propto e^{-\beta \frac{l_p}{2} \phi / ds}

    or just a Gaussian, with variance given by :math:`ds/l_p`, the number of
    persistence lengths between the beads.

    Recall the shape of the sphere means that correctly integrating over the
    uniform distribution in the spherical angle leaves us with

    .. math::

        P(\phi) \propto \sin(\phi) e^{-\beta \frac{l_p}{2} \phi / ds}

    the normalization factor to make this a probability distribution for a
    given :math:`ds/l_p` can be calculating analytically (restricting
    explicitly to :math:`\phi\in[0,\pi]` makes the formula uglier but
    WolframAlpha can still do it), yielding an exact probability distribution

    .. math::
        P(\phi) = \sqrt{\frac{2}{\pi (ds/l_p)}} \frac{-1}{e^{ds/2l_p}
        \erf{\sqrt{ds/2l_p}}} \sin(\phi) e^{phi^2/2(ds/l_p)}

    which we can use to inverse-transform sample phi, then draw theta
    uniformly.

    Given theta, phi, we can then get the next tangent vector (and thus next
    location, since for discrete WLC we define u = diff(r)) from the current
    one.

    Notes
    -----
    We using the formalism above, we can get an estimate for what values of
    :math:`ds/l_p` are reasonable to use. For example, for eps=0.1,
    we will typically see values of phi less than pi/3. for eps=0.01, we see
    phi less than phi/10 or so.
    """
    l0 = L/(N-1)
    eps = lp/l0 # ds

    r = np.zeros((N,3))
    Omega = np.identity(3) # n, b, u vectors, arranged in columns
    r[1] = l0*Omega[:,2] # initial in z-direction

    # pre-generate the random numbers for speed
    thetas = 2*np.pi*urand(size=(N-2,))
    xis = urand(size=(N-2,))
    f0, fpi = phi_Z_0_1_(eps)
    with Pool(n_proc) as p:
        phis = p.map(partial(inv_sample_, eps=eps, f0=f0, fpi=fpi), xis)

    for j in range(2, N):
        # first rotate about tangent vector by theta, then "bend" by rotating
        # about the new "normal" vector
        Omega = Omega @ Rz(thetas[j-2]) @ Ry(phis[j-2])

        # third column of orientation matrix is new tangent vector
        r[j] = r[j-1] + l0*Omega[:,2]

    return r

def dwlc(N, L, lp):
    pass

def wlc_init(N, L, lp, lt=0):
    fortran_code = """
subroutine wlc_init(R, U, NB, EPS, l0, rand_stat)
        ! takes R(3,NB) with R(:,1) preset and makes a WLC given EPS
    use mersenne_twister, only : random_number, random_stat
    use params, only : dp, pi

    implicit none

    integer, intent(in) :: NB
    real(dp), intent(in) :: EPS, l0
    real(dp), intent(inout) :: R(3,NB), U(3,NB)
    type(random_stat), intent(inout) :: rand_stat

    integer J
    real(dp) N1(3), N2(3), z, theta
    real(dp) urand(3)

    do J = 2,NB

         call random_number(urand,rand_stat)
         theta = urand(1)*2.0_dp*pi
         z = (1.0_dp/EPS)*log(2.0_dp*sinh(EPS)*urand(2)+exp(-EPS))

         N1 = (/ 0.0_dp, 0.0_dp, 1.0_dp /)
         N1 = N1 - dot_product(N1, U(:,J-1))*U(:,J-1)
         N1 = N1/norm2(N1)

         N2 = cross(N1, U(:,J-1))
         N2 = N2/norm2(N2)

         U(:,J) = sqrt(1-z*z)*(cos(theta)*N1 + sin(theta)*N2 + z*U(:,J-1))
         U(:,J) = U(:,J)/norm2(U(:,J))

         if (WLC_P__LOCAL_TWIST) then
             print*, "wlc chain initialization is not implimented for local twist"
             stop
         endif

         R(:,J) = R(:,J-1) + l0*U(:,J)
     enddo
end subroutine wlc_init
    """
    l0 = L/(N-1)
    eps = l0/lp

    r = np.zeros((N,3))
    u = np.zeros((N,3))
    u[:,2] = 1
    n = np.zeros((N,3))
    n[:,1] = 1
    for j in range(1, N):
        # first get an arbitrary basis n1, n2 to the cotangent plane to u[j-1]
        n1 = np.array([0, 0, 1])
        n1 = n1 - (n1@u[j-1])*u[j-1]
        mag = np.linalg.norm(n1)
        # if we accidentally chose vector parallel to u[j-1]
        if np.isclose(mag, 0):
            n1 = np.array([0, 1, 0])
            n1 = n1 - (n1@u[j-1])*u[j-1]
            mag = np.linalg.norm(n1)
        n1 = n1/mag
        n2 = np.cross(n1, u[j-1])
        n2 = n2/np.linalg.norm(n2)

        # now at a random angle in cotangent plane
        theta = 2*np.pi*urand(1)
        # WLC formula for the magnitude of u[j]@u[j-1]
        z = (1/eps)*np.log(2*urand()*np.sinh(eps)+np.exp(-eps))
        u[j] = np.sqrt(1-z*z)*(np.cos(theta)*n1 + np.sin(theta)*n2 + z*u[j-1])
        u[j] = u[j]/np.linalg.norm(u[j])

        r[j] = r[j-1] + l0*u[j]
    return r, u, n


def effective_wormlike_chain_init(N, L, lp, lt):
    """Generate a wormlike chain with potentially less beads than is wise for
    direct sampling of WLC with 'quadratic' approximation used by wlc
    function, by "regridding" more finely then only saving the beads that were
    originally requested."""

    fortran_code = """subroutine effective_wormlike_chain_init(R, U, NT, wlc_p, rand_stat)
    use mersenne_twister
    use params, only : wlcsim_params, dp, max_wlc_l0, maxWlcDelta
    use vector_utils, only: randomUnitVec
    use polydispersity, only: length_of_chain
    implicit none
    integer, intent(in) :: nt
    type(wlcsim_params), intent(in) :: wlc_p
    real(dp), intent(out) :: R(3,nt), U(3,nt)
    type(random_stat), intent(inout) :: rand_stat

    integer IB, NgB, i, j
    real(dp) l0, eps
    real(dp) urand(3)
    real(dp), dimension(:,:), allocatable :: tmpR, tmpU

    IB = 1
    if (wlc_p%DEL > max_wlc_l0) then
        NgB = ceiling(wlc_p%DEL/max_wlc_l0) + 1
    else
        NgB = 1 + 1
    endif
    allocate(tmpR(3,NgB))
    allocate(tmpU(3,NgB))
    l0 = wlc_p%DEL/(NgB - 1)
    EPS = WLC_P__LP/l0 ! bending rigidity for wormlike chain
    do I = 1,WLC_P__NP
        ! uniformly first bead inside box
        call random_number(urand,rand_stat)
        R(1,IB) = urand(1)*WLC_P__LBOX_X
        R(2,IB) = urand(2)*WLC_P__LBOX_Y
        R(3,IB) = urand(3)*WLC_P__LBOX_Z
        ! uniformly from unit sphere first tan vec
        call randomUnitVec(U(:,IB),rand_stat)
        IB = IB + 1
        do J = 2,length_of_chain(I)
            tmpR(:,1) = R(:,IB-1)
            tmpU(:,1) = U(:,IB-1)
            call wlc_init(tmpR, tmpU, NgB, EPS, l0, rand_stat)
            R(:,IB) = tmpR(:,NgB)
            U(:,IB) = tmpU(:,NgB)
            IB = IB + 1
        enddo
    enddo
    deallocate(tmpR)
    deallocate(tmpU)
end subroutine effective_wormlike_chain_init
"""




import re
import os
import subprocess

import pandas as pd
import numpy as np

self_dir, self_file = os.path.split(os.path.realpath(__file__))
src_dir = os.path.join(self_dir, '..', 'src')

mc_moves = pd.Series({
    1: 'crank_shaft',
    2: 'slide_move',
    3: 'pivot_move',
    4: 'rotate_move',
    5: 'full_chain_rotation',
    6: 'full_chain_slide',
    7: 'change_binding_state',
    8: 'chain_flip',
    9: 'chain_exchange',
    10: 'reptation',
    11: 'super_reptation'
})
xyz_names = pd.Series({1: 'x', 2: 'y', 3: 'z'})

#####{{{
# parsing functions

var_re = "[a-zA-Z0-9][_a-zA-Z0-9]*"
float_re = "[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"

def extract_from_lines(lines, object_type):
    """Return a DataFrame containing the appropriate columns for the requested
    object to be extracted linewise from a list of lines.

    Objects are parsed only into their text representations, like a
    pre-processor. Ask __dict__.object_types for valid objects to request."""
    object_types = {'declaration': declaration,
                    'wlc_p_default': wlc_p_default}
    if object_type not in object_types:
        raise ValueError(str(object_type) + ' should be one of the strings: '
                         + ', '.join(object_types.keys()))
    objects = [object_types[object_type](line) for line in lines]
    parsed = pd.DataFrame([obj for obj in objects if obj is not None])
    parsed.columns = object_types[object_type].column_names
    return parsed

def declaration(line):
    line_re = re.compile("""
        \s*                                 # leading whitespace
        (real|integer|character|logical)    # type
        (\((""" + var_re + """)\))?         # optional "precision", int or varname
        \s+                                 # some whitespace
        ([_0-9a-zA-Z]+)                     # variable name
        (\((""" + var_re + """)\))?         # shape of the variable, opt.
        \s*(!.*)?                           # optional following comment
                          """, re.VERBOSE)
    match = line_re.search(line)
    if match is None:
        return None
    # some groupings used for optionality, these are ignored
    type, _, type_size, name, _, shape, comment = match.groups()
    return type, type_size, name, shape, comment
declaration.column_names = ['type', 'type_size', 'name', 'shape', 'comment']

def wlc_p_default(line):
    line_re = re.compile("""
        \s*                     # leading whitespace
        wlc_p%(""" + var_re + """)   # variable name
        (\((""" + var_re + """)\))?  # indexer, capture also w/o parens
        \s*=\s*
        ([^\s!]+)               # value, terminated by whitespace or a comment
        \s*(!.*)?               # optional following comment
                                """, re.VERBOSE)
    match = line_re.search(line)
    if not match:
        return None
    name, _, indexer, value, comment = match.groups()
    return name, indexer, value, comment
wlc_p_default.column_names = ['name', 'indexer', 'value', 'comment']

# end parsing functions
#####}}}

#####{{{
# extract parts of the code

def get_function_body(function_name, file):
    function_lines = []
    with open(file) as f:
        start_re = re.compile('subroutine\s' + function_name)
        end_re = re.compile('end subroutine')
        appending = False
        for line in f:
            if start_re.search(line):
                appending = True
                continue
            if end_re.search(line):
                break
            if appending:
                function_lines.append(line)
    return function_lines

def get_type_body(type_name, file):
    type_lines = []
    with open(file) as f:
        start_re = re.compile('type\s' + type_name)
        end_re = re.compile('end type')
        appending = False
        for line in f:
            if start_re.search(line):
                appending = True
                continue
            if end_re.search(line):
                break
            if appending:
                type_lines.append(line)
    return type_lines

# end extract parts of the code
#####}}}


#####{{{
# do big code modifications

def wlc_p_to_defines():
    """gfortran is apparently bad at expression folding if there's not constants
    involved. So instead of precomputing all divisions by hand, we make all
    parameters compile time constants to try to offload this job into the
    compiler's constant folding step.

    This function extracts the wlcsim_params type, gets its default values from
    params.set_param_defaults, making a DataFrame with all the information
    needed for make_defines_inc to do it's job.

    Special care must be taken with variables like minWindow. These are used as
    arrays, one index for each move. We keep the defaults lines for these
    variables, unrolling the for loops manually to have the defaults be like
    ```
    wlc_p%minWindow(2) = wlc_p__minWindow_slide
    ```

    Some variables are kept regardless, since they are defined by reading in
    from a table that Elena generated or depend in some "turing-complete" way
    on other input parameters.
    """
    vars_to_keep = ['EB', 'EPAR', 'EPERP', 'ESELF', 'GAM', 'ETA', 'XIR',
            'SIGMA', 'XIU', 'DEL', 'LHC', 'VHC', 'REND', 'L0', 'EPS',
            'DT', 'NBIN', 'NT', 'SIMTYPE',
                   # parallel tempered variables, quinn
            'CHI', 'MU', 'HA', 'HP1_BIND', 'KAP', 'CHI_L2',
            'RECENTER_ON', 'KAP_ON', 'CHI_ON', 'COUPLE_ON', 'FIELD_INT_ON',
            'BIND_ON', "CHI_L2_ON",
                   # parallel tempered variables, brad
            'LK' ]
    wlc_p_usage_re = re.compile("[^!]*wlc_p%("+ var_re +")\(("+ var_re +")\)")
    wlc_p_usage_re_grep =      "^[^!]*wlc_p%" + var_re + "\(" + var_re + "\)"
    defaults = extract_from_lines(
            get_function_body('set_param_defaults', 'src/wlcsim/params.f03'),
            'wlc_p_default')
    for col in ['name', 'indexer', 'value']:
        defaults[col] = defaults[col].str.upper()
    variables = extract_from_lines(
            get_type_body('wlcsim_params', 'src/wlcsim/params.f03'),
            'declaration')
    for col in ['type', 'type_size', 'name', 'shape']:
        variables[col] = variables[col].str.upper()
    array_usages = subprocess.check_output(['grep', '-ohER',
            wlc_p_usage_re_grep, src_dir]).splitlines()
    array_usages = [wlc_p_usage_re.search(str(u)) for u in array_usages]
    array_usages = (pd.DataFrame(
            [m.groups() for m in array_usages if m is not None])
            .applymap(str.upper) # "parse" into caps, fortran case-insensitive
            .drop_duplicates())  # get unique name,indexer pairs
    array_usages.columns = ['name', 'indexer']
    is_looped_on = array_usages.indexer.str.contains('[A-Z]')
    looped_names = array_usages.name[is_looped_on]

    defines = []
    defines_columns = ['define_name', 'name', 'indexer', 'value',
                       'declaration_comment', 'defaults_comment',
                       'type', 'type_size', 'shape', 'is_looped_on',
                       'keep_regardless']
    # not_found = []
    for variable in variables.itertuples():
        if variable.shape is None:
            define_name = 'WLC_P__' + variable.name
            var_ix = defaults.name == variable.name
            if variable.name in vars_to_keep:
                value = None
            elif not var_ix.any():
                raise ValueError('Could not find default for variable ' + variable.name)
                # not_found.append(variable.name)
            else:
                value = defaults.value[var_ix].values[0]
            if value is None:
                dcom = None
            else:
                dcom = defaults.comment[var_ix].values[0]
            defines.append((define_name, variable.name, None,
                    value, variable.comment, dcom, variable.type,
                    variable.type_size, variable.shape,
                    variable.name in looped_names.values,
                    variable.name in vars_to_keep))
            continue
        if variable.shape == 'NMOVETYPES':
            name_map = mc_moves
        elif variable.shape == 'NDIM':
            name_map = xyz_names
        else:
            name_map = {i: str(i) for i in range(1, int(variable.shape)+1)}
        for idx, name in name_map.items():
            define_name = 'WLC_P__' + variable.name + '_' + name.upper()
            indexer = str(idx)
            var_ix = (defaults.name == variable.name) & (defaults.indexer == indexer)
            # it's possible that the variable is set in a loop to the same
            # value
            if not var_ix.any():
                var_ix = (defaults.name == variable.name) & defaults.indexer.str.contains('[A-Z]')
            # that it's set using vector syntax
            if not var_ix.any():
                var_ix = (defaults.name == variable.name) & defaults.indexer.isnull()
            # but all variables should have a default, even if it's a default
            # that forces an error, unless they're "derived"
            if variable.name in vars_to_keep:
                value = None
            elif not var_ix.any():
                raise ValueError('Could not find default for variable ' + variable.name)
                # not_found.append(variable.name)
            else:
                value = defaults.value[var_ix].values[0]
            dcom = defaults.comment[var_ix].values[0]
            defines.append((define_name, variable.name, indexer,
                    value, variable.comment, dcom, variable.type,
                    variable.type_size, variable.shape,
                    variable.name in looped_names.values,
                    variable.name in vars_to_keep))
    defines = pd.DataFrame(defines)
    defines.columns = defines_columns
    return defines

def make_defines_inc(active_defines):
    """Takes defines from wlc_p_to_defines,

    deletes unused variables from the wlcsim_params type and from the defaults
    function, creates a 'defines.inc' file, and replaces all uses of 'wlc_p%NAME'
    with 'wlc_p__NAME' by brute force regex.
    """
    defines_file = os.path.join(src_dir, 'defines.inc')
    if os.path.exists(defines_file):
        raise OSError(defines_file + ' already exists!')
    with open(defines_file, 'w') as f:
        # f.write("\n\n#define PI 4*atan(1.0_dp)\n\n")
        for new_define in active_defines.itertuples():
            f.write("! VARIABLE COMMENT: {}\n".format(new_define.declaration_comment))
            f.write("! DEFAULTS COMMENT: {}\n".format(new_define.defaults_comment))
            if new_define.type_size:
                f.write("! TYPE: {}({})\n".format(new_define.type, new_define.type_size))
            else:
                f.write("! TYPE: {}\n".format(new_define.type))
            f.write("#define {} {}\n".format(new_define.define_name, new_define.value))

def temp_rename(active_defines, undo=False):
    names_in_replace_order = get_replace_order(active_defines.name.values)
    # now replace all instances of wlc_p variables with their new values
    for name in names_in_replace_order:
        if undo:
            program = ['find', src_dir, '-type', 'f', '-exec', 'sed', '-i',
                    's/TMPTMPTMP' + name + '/wlc_p%' + name+ '/gi',
                    '{}', ';']
        else:
            program = ['find', src_dir, '-type', 'f', '-exec', 'sed', '-i',
                    's/wlc_p%' + name + '/TMPTMPTMP' + name + '/gi',
                    '{}', ';']
        subprocess.check_output(program)

def replace_wlc_p_instances(active_defines):
    # def shape_to_int(shape):
    #     if shape == 'NDIM':
    #         return len(xyz_names)
    #     elif shape == 'NMOVETYPES':
    #         return len(mc_moves)
    #     else:
    #         return int(shape)
    affected_files = subprocess.check_output(['grep', '-iRl', 'wlc_p%', src_dir]).splitlines()
    for file in affected_files:
        # if os.path.basename(file) == b'defines.inc':
        #     continue
        if not re.search("\.f[0-9][0-9]'$", str(file)) and not re.search("\.inc'$", str(file)):
            continue
        # add the include line at the top for defines.inc
        if os.path.basename(file) != b"defines.inc":
            with open(file, 'r') as original: data = original.read()
            with open(file, 'w') as modified: modified.write('#include "../defines.inc"\n' + data)
        names_in_replace_order = get_replace_order(active_defines.name.values)
        # now replace all instances of wlc_p variables with their new values
        for name in names_in_replace_order:
            defines_with_name = active_defines[active_defines.name == name]
            for define in defines_with_name.itertuples():
                if define.shape is None:
                    subprocess.check_output(['sed', '-i',
                            's/wlc_p%' + define.name + '/' + define.define_name + '/gi',
                            file])
                else:
                    subprocess.check_output(['sed', '-i',
                            's/wlc_p%' + define.name + '(' + define.indexer + ')/'
                            + define.define_name + '/gi', file])

def get_replace_order(strings):
    """Return a list of tuples corresponding to the defines that we are going to
    replace, but in order such that all superstrings are replaced first."""
    lengths = np.array([-len(string) for string in strings])
    ix = np.argsort(lengths)
    return np.array(strings)[ix]

def perform_wlc_p_to_define_transform():
    defines = wlc_p_to_defines()
    to_delete = defines['shape'].isnull() & (~defines.keep_regardless)
    declarations_to_delete = defines[to_delete]
    declarations_to_keep = defines[~to_delete]
    delete_declaration_lines(declarations_to_delete.name.values, os.path.join(src_dir, 'wlcsim', 'params.f03'),
                             in_type='wlcsim_params')
    # delete_definition_lines(defines.name.values, os.path.join(src_dir, 'wlcsim', 'params.f03'),
    #                         in_subroutine='set_param_defaults')
    delete_subroutine('set_param_defaults', os.path.join(src_dir, 'wlcsim', 'params.f03'))
    delete_subroutine('read_input_file', os.path.join(src_dir, 'wlcsim', 'params.f03'))
    defines_to_be_written = defines.loc[~defines.keep_regardless]
    make_defines_inc(defines_to_be_written)
    temp_rename(declarations_to_keep)
    replace_wlc_p_instances(declarations_to_delete)
    temp_rename(declarations_to_keep, undo=True)
    add_looped_defaults(defines)

def delete_subroutine(name, file):
    with open(file, 'r') as f:
        lines = f.readlines()
    with open(file, 'w') as f:
        pos = "before subroutine"
        for line in lines:
            if pos == "before subroutine" and re.search("subroutine\s+" + name, line):
                    pos = "inside subroutine"
            elif pos == "inside subroutine" and re.search("end\s+subroutine", line):
                    pos = "after subroutine"
                    continue
            if pos != 'inside subroutine':
                f.write(line)

def add_looped_defaults(defines):
    params_file = os.path.join(src_dir, 'wlcsim', 'params.f03')
    with open(params_file, 'r') as f:
        old_file = f.readlines()
    with open(params_file, 'w') as f:
        pos = 'before module'
        for line in old_file:
            if pos == 'before module' and re.search('module\s+params', line):
                pos = 'before contains'
            if pos == 'before contains' and re.match('^\s*contains\s*$', line):
                pos = 'after contains'
                f.write(line)
                write_looped_defaults(f, defines)
                continue
            f.write(line)

def write_looped_defaults(f, defines):
    f.write("""
    subroutine set_param_defaults(wlc_p)
        implicit none
        ! WARNinG: changing this to intent(out) means that unassigned values
        ! here will become undefined upon return, due to Fortran's weird
        ! intent(out) semantics for records, this would require that a default
        ! value always be given to new parameters in wlc_p, else we would get a
        ! compile time catchable runtime error that is not caught by gcc as of v5.0
        !
        ! this is almost definitely undesireable, since "undefined" means the
        ! behavior will depend on which compiler is used
        type(wlcsim_params), intent(inout) :: wlc_p
""")
    for define in defines.itertuples():
        if define.indexer is None:
            continue
        f.write("        wlc_p%{}({}) = {}\n".format(
                define.name, define.indexer, define.define_name))
    f.write("""
    end subroutine set_param_defaults

""")

def delete_declaration_lines(names, file, in_type=None):
    with open(file, 'r') as f:
        lines = f.readlines()
    with open(file, 'w') as f:
        pos = "before type"
        for line in lines:
            if in_type and pos == "before type":
                if re.search("type\s+" + in_type, line):
                    pos = "inside type"
            elif in_type and pos == "inside type":
                if re.search("end\s+type", line):
                    pos = "after type"
            if not in_type or pos == "inside type":
                match = declaration(line)
                if match and match[2].upper() in names:
                    continue
            f.write(line)

def delete_definition_lines(names, file, in_subroutine=None):
    with open(file, 'r') as f:
        lines = f.readlines()
    with open(file, 'w') as f:
        pos = "before subroutine"
        for line in lines:
            if in_subroutine and pos == "before subroutine":
                if re.search("subroutine\s+" + in_subroutine, line):
                    pos = "inside subroutine"
            elif in_subroutine and pos == "inside subroutine":
                if re.search("end\s+subroutine", line):
                    pos = "after subroutine"
            if not in_subroutine or pos == "inside subroutine":
                match = wlc_p_default(line)
                if match and match[0].upper() in names:
                    continue
            f.write(line)

def defines_to_wlc_p():
    """The inverse of wlc_p_to_defines. //TODO"""
    #TODO implement
    pass



# end do big code modifications
#####}}}
